import unittest
from unittest.mock import MagicMock

from machine_common_sense.config_manager import (ConfigManager,
                                                 SceneConfiguration)
from machine_common_sense.controller_events import (AfterStepPayload,
                                                    BeforeStepPayload,
                                                    EndScenePayload,
                                                    StartScenePayload)
from machine_common_sense.history_writer import (HistoryEventHandler,
                                                 HistoryWriter, SceneHistory)

TEST_FILE_NAME = "test_scene_file.json"


class TestHistoryEventHandler(unittest.TestCase):

    scene_config = SceneConfiguration(name=TEST_FILE_NAME)
    config_mngr = ConfigManager()

    def setUp(self):
        self.config_mngr._config[
            ConfigManager.CONFIG_DEFAULT_SECTION
        ] = {
            ConfigManager.CONFIG_EVALUATION_NAME: "test-name",
            ConfigManager.CONFIG_HISTORY_ENABLED: True,
            ConfigManager.CONFIG_METADATA_TIER: "level1",
        }

        self.histEvents = HistoryEventHandler()

    def test_on_start_scene(self):
        test_payload = {
            "step_number": 1,
            "config": self.config_mngr,
            "scene_config": self.scene_config,
            "output_folder": None,
            "timestamp": "20210831-202203",
            "wrapped_step": {},
            "step_metadata": {},
            "step_output": {},
            "restricted_step_output": {},
            "goal": {}
        }

        self.assertIsNone(
            self.histEvents._HistoryEventHandler__history_writer)

        self.histEvents.on_start_scene(StartScenePayload(**test_payload))

        self.assertIsNotNone(
            self.histEvents._HistoryEventHandler__history_writer)
        self.assertIsNotNone(
            self.histEvents._HistoryEventHandler__history_writer)

    def test_on_start_scene_hist_not_enabled(self):
        self.config_mngr._config[
            ConfigManager.CONFIG_DEFAULT_SECTION
        ] = {
            ConfigManager.CONFIG_EVALUATION_NAME: "test-name",
            ConfigManager.CONFIG_HISTORY_ENABLED: False,
            ConfigManager.CONFIG_METADATA_TIER: "level1",
        }
        test_payload = {
            "step_number": 0,
            "config": self.config_mngr,
            "scene_config": self.scene_config,
            "output_folder": None,
            "timestamp": "20210831-202203",
            "wrapped_step": {},
            "step_metadata": {},
            "step_output": {},
            "restricted_step_output": {},
            "goal": {}
        }

        self.assertIsNone(
            self.histEvents._HistoryEventHandler__history_writer)

        self.histEvents.on_start_scene(StartScenePayload(**test_payload))

        self.assertIsNone(
            self.histEvents._HistoryEventHandler__history_writer)

    def test_on_before_step_zero(self):
        self.histEvents._HistoryEventHandler__history_writer = MagicMock()
        test_payload = {
            "step_number": 0,
            "config": self.config_mngr,
            "scene_config": self.scene_config,
            "action": "Initialize",
            "habituation_trial": None,
            "goal": {}
        }

        self.histEvents.on_before_step(BeforeStepPayload(**test_payload))

        hist_writer = self.histEvents._HistoryEventHandler__history_writer

        hist_writer.init_timer.assert_called()
        hist_writer.add_step.assert_not_called()

    def test_on_before_step_one(self):
        histItem = SceneHistory(
            step=1,
            action="MoveAhead",
            args=None,
            params=None,
            output=None,
            delta_time_millis=0)
        self.histEvents._HistoryEventHandler__history_item = histItem
        self.histEvents._HistoryEventHandler__history_writer = MagicMock()
        test_payload = {
            "step_number": 1,
            "config": self.config_mngr,
            "scene_config": self.scene_config,
            "action": "Initialize",
            "habituation_trial": None,
            "goal": {}
        }

        self.histEvents.on_before_step(BeforeStepPayload(**test_payload))

        hist_writer = self.histEvents._HistoryEventHandler__history_writer

        hist_writer.init_timer.assert_not_called()
        hist_writer.add_step.assert_called_with(histItem)

    def test_on_after_step(self):
        test_payload = {
            "step_number": 1,
            "ai2thor_action": "MoveAhead",
            "action_kwargs": {},
            "step_params": {},
            "config": self.config_mngr,
            "scene_config": self.scene_config,
            "goal": {},
            "output_folder": None,
            "timestamp": "20210831-202203",
            "wrapped_step": {},
            "step_metadata": {},
            "step_output": MagicMock(),
            "restricted_step_output": {}
        }

        after_step_payload = AfterStepPayload(**test_payload)

        self.histEvents.on_after_step(after_step_payload)

        histItem = self.histEvents._HistoryEventHandler__history_item

        after_step_payload.step_output.copy_without_depth_or_images.assert_called()  # noqa: E501
        self.assertEqual(histItem.step, 1)
        self.assertEqual(histItem.action, "MoveAhead")
        self.assertEqual(histItem.delta_time_millis, 0)

    def test_on_end_scene(self):
        self.histEvents._HistoryEventHandler__history_writer = HistoryWriter(
            self.scene_config, {}, "20210831-202203")
        self.histEvents._HistoryEventHandler__history_writer.write_history_file = MagicMock()  # noqa: E501
        histItem = SceneHistory(
            step=1,
            action="MoveAhead",
            args=None,
            params=None,
            output=None,
            delta_time_millis=0)
        self.histEvents._HistoryEventHandler__history_item = histItem
        test_payload = {}
        test_payload["step_number"] = 1
        test_payload["config"] = self.config_mngr
        test_payload["scene_config"] = self.scene_config
        test_payload["rating"] = "plausible"
        test_payload["score"] = .8
        test_payload["report"] = {1: {"rating": "plausible",
                                      "score": .75,
                                      "violations_xy_list": [
                                          {
                                              "x": 1,
                                              "y": 1
                                          }
                                      ],
                                      "internal_state": {
                                          "test": "some state"
                                      }}
                                  }

        self.histEvents.on_end_scene(EndScenePayload(**test_payload))

        hist_writer = self.histEvents._HistoryEventHandler__history_writer
        self.assertEqual(len(hist_writer.current_steps), 1)
        self.assertEqual(
            hist_writer.current_steps[0]["action"], "MoveAhead")
        self.assertEqual(
            hist_writer.current_steps[0]["classification"],
            "plausible")
        self.assertEqual(
            hist_writer.current_steps[0]["confidence"],
            0.75)
        self.assertEqual(
            hist_writer.current_steps[0]["violations_xy_list"],
            [
                {
                    "x": 1,
                    "y": 1
                }
            ])
        self.assertEqual(
            hist_writer.current_steps[0]["internal_state"],
            {
                "test": "some state"
            })
        self.assertEqual(
            hist_writer.current_steps[0]["internal_state"],
            {
                "test": "some state"
            })

        hist_writer.write_history_file.assert_called_with("plausible", .8)

    def test_on_end_scene_history_not_enabled(self):
        self.config_mngr._config[
            ConfigManager.CONFIG_DEFAULT_SECTION
        ] = {
            ConfigManager.CONFIG_EVALUATION_NAME: "test-name",
            ConfigManager.CONFIG_HISTORY_ENABLED: False,
            ConfigManager.CONFIG_METADATA_TIER: "level1",
        }
        self.histEvents._HistoryEventHandler__history_writer = MagicMock()  # noqa: E501

        test_payload = {}
        test_payload["step_number"] = 1
        test_payload["config"] = self.config_mngr
        test_payload["scene_config"] = self.scene_config
        test_payload["rating"] = "plausible"
        test_payload["score"] = .8
        test_payload["report"] = {1: {"rating": "plausible",
                                      "score": .75,
                                      "violations_xy_list": [
                                          {
                                              "x": 1,
                                              "y": 1
                                          }
                                      ],
                                      "internal_state": {
                                          "test": "some state"
                                      }}
                                  }

        self.histEvents.on_end_scene(EndScenePayload(**test_payload))

        hist_writer = self.histEvents._HistoryEventHandler__history_writer
        hist_writer.add_step.assert_not_called()
        hist_writer.write_history_file.assert_not_called()


if __name__ == '__main__':
    unittest.main()



from enum import Enum, unique

from .action import Action
from .stringifier import Stringifier


class GoalMetadata:
    """
    Defines metadata for a goal in the MCS 3D environment.

    Attributes
    ----------
    action_list : list of lists of (string, dict) tuples, or None
        The list of all actions that are available for the scene at each step
        (outer list). Each inner list is the list of all actions that are
        available for the single step corresponding to the inner list's index
        within the outer list. Each action is returned as a tuple containing
        the action string and the action's restricted paramters, if any.
        For example: ("Pass", {}) forces a Pass action; ("PickupObject", {})
        forces a PickupObject action with any parameters; and
        ("PickupObject", {"objectId": "a"}) forces a PickupObject action with
        the specific parameters objectId=a.
        An action_list of None means that all actions are always available.
        An empty inner list means that all actions will be available on that
        specific step.
        See StepMetadata.action_list for the available actions of the current
        step.
        May be a subset of all possible actions. See
        :mod:`Action <machine_common_sense.Action>`.
    category : string
        The category that describes this goal and the properties in its
        metadata. See :mod:`Goal <machine_common_sense.GoalCategory>`.
    description : string
        A human-readable sentence describing this goal and containing
        the target task(s) and object(s).

        Sizes:
        - tiny: near the size of a baseball
        - small: near the size of a baby
        - medium: near the size of a child
        - large: near the size of an adult
        - huge: near the size of a sofa

        Weights:
        - light: can be held by a baby
        - heavy: cannot be held by a baby, but can be pushed or pulled
        - massive: cannot be moved by a baby

        Colors:
        black, blue, brown, green, grey, orange, purple, red, white, yellow

        Materials:
        See :mod:`Material <machine_common_sense.Material>`.
    habituation_total : int
        The total count of habituation trials that will be in this scene.
    last_preview_phase_step : integer
        The last step of the Preview Phase of this scene, if a Preview Phase is
        scripted in the scene configuration. Each step of a Preview Phase
        normally has a single specific action defined in this goal's
        action_list property for the performer to choose, like ['Pass'].
        Default: 0 (no Preview Phase)
    last_step : integer
        The last step of this scene. This scene will automatically end
        following this step.
    metadata : dict
        The metadata specific to this goal. See
        :mod:`Goal <machine_common_sense.GoalCategory>`.
    """

    ACTION_LIST = [(item.value, {}) for item in Action]

    def __init__(
        self,
        action_list=None,
        category='',
        description='',
        habituation_total=0,
        last_preview_phase_step=0,
        last_step=None,
        metadata=None
    ):
        # The action_list must be None by default
        self.action_list = action_list
        self.category = category
        self.description = description
        self.habituation_total = habituation_total
        self.last_preview_phase_step = last_preview_phase_step
        self.last_step = last_step
        self.metadata = {} if metadata is None else metadata

    def __str__(self):
        return Stringifier.class_to_str(self)

    # Allows converting the class to a dictionary, along with allowing
    #   certain fields to be left out of output file
    def __iter__(self):
        yield 'action_list', self.action_list
        yield 'category', self.category
        yield 'description', self.description
        # yield 'domain_list', self.domain_list
        # yield 'info_list', self.info_list
        yield 'last_preview_phase_step', self.last_preview_phase_step
        yield 'last_step', self.last_step
        # yield 'type_list', self.type_list
        yield 'metadata', self.metadata

    def retrieve_action_list_at_step(self, step_number):
        """Return the action list from the given goal at the given step as a
        a list of actions tuples by default."""
        if self is not None and self.action_list is not None:
            if step_number < self.last_preview_phase_step:
                return ['Pass']
            if self.last_step is not None and step_number == self.last_step:
                return []
            adjusted_step = step_number - self.last_preview_phase_step
            if len(self.action_list) > adjusted_step:
                if len(self.action_list[adjusted_step]) > 0:
                    return self.action_list[adjusted_step]

        return GoalMetadata.ACTION_LIST


@unique
class GoalCategory(Enum):
    """
    Each goal dict will have a "category" string that describes the type of
    scene (or, the type of task within the scene) being run. Each goal dict
    will also have a "metadata" dict containing one or more properties
    depending on the "category".
    """

    AGENTS = "agents"
    """
    In a trial that has an Agents goal, you must sit and observe a scene as one
    or more simulation-controlled agents act in predefined ways within your
    camera's viewport, and then decide whether the scene is "expected" or
    "unexpected". The camera will always be positioned at an isometric
    perspective, like you're standing on an elevated platform looking down at
    the scene. Each scene will consist of eight sequential habituation trials,
    depicting expected agent behaviors and separated by EndHabituation actions
    (each of which generates a black frame image when called), immediately
    followed by the test trial, depicting either an expected or unexpected
    agent behavior. All nine of these trials happen within the same "scene".
    These trials will demand a "common sense" understanding of agents, their
    behaviors, and their interactions with objects in the environment.

    Notes
    -----
    You are required to call `controller.end_scene()` at the end of each scene
    with a continuous plausibility `rating`, from 0.0 (completely implausible)
    to 1.0 (completely plausible). You are not required to also pass it a
    `score`.
    """

    INTUITIVE_PHYSICS = "intuitive physics"
    """
    In a trial that has an Intuitive Physics goal, you must sit and observe a
    scene as objects move across your camera's viewport, and then decide
    whether the scene is "plausible" or "implausible". These trials will demand
    a "common sense" understanding of basic ("intuitive") physics, like object
    permanence or shape constancy. Inspired by Emmanuel Dupoux's "IntPhys: A
    Benchmark for Visual Intuitive Physics Reasoning" (http://intphys.com).

    Notes
    -----
    You are required to call `controller.end_scene()` at the end of each scene
    with a binary plausibility `rating` -- either 0 (implausible) or 1
    (plausible) -- and a continuous plausibility `score` -- from 0.0
    (completely implausible) to 1.0 (completely plausible). This is also
    where you would submit any retrospective reporting on a per step basis via
    `report`.
    """

    RETRIEVAL = "retrieval"
    """
    In a trial that has a retrieval goal, you must find and pickup a target
    object. This may involve exploring the scene, avoiding obstacles,
    interacting with objects (like closed containers), and (future evaluations)
    tracking moving objects. These trials will demand a "common sense"
    understanding of self navigation (how to move and rotate yourself within a
    scene and around obstacles), object interaction (how objects work,
    including opening containers), and (future evaluations) the basic physics
    of movement (kinematics, gravity, friction, etc.).

    Parameters
    ----------
    target.id : string
        The objectId of the target object to retrieve.
        Will only be available at `oracle` metadata level.

    target.info : list of strings
        Human-readable information describing the target object needed for the
        visualization interface.

    """

    TRANSFERRAL = "transferral"
    """
    NOT USED IN MCS EVAL 4+

    In a trial that has a transferral goal, you must find and pickup the
    first target object and put it down either next to or on top of the second
    target object. This may involve exploring the scene, avoiding obstacles,
    interacting with objects (like closed receptacles), and (future
    evaluations) tracking moving objects. These trials will demand a "common
    sense" understanding of self navigation (how to move and rotate yourself
    within a scene and around obstacles), object interaction (how objects work,
    including opening containers), and (future evaluations) the basic physics
    of movement (kinematics, gravity, friction, etc.).

    Parameters
    ----------
    relationship : list of strings
        The required final position of the two target objects in relation to
        one another. For transferral goals, this value will always be either
        ["target_1", "next_to", "target_2"] or ["target_1", "on_top_of",
        "target_2"].

    target_1.id : string
        The objectId of the first target object to pickup and transfer to the
        second target object.

    target_1.image : list of numpy arrays
        An image of the first target object to pickup and transfer to the
        second target object, given as a 3D RGB pixel array.

    target_1.info : list of strings
        Human-readable information describing the target object needed for the
        visualization interface.

    target_1.match_image : string
        Whether the image of the first target object (target_1.image) exactly
        matches the actual object in the scene. If false, then the actual first
        target object will be different in one way (for example, the image may
        depict a blue ball, but the actual object is a yellow ball, or a blue
        cube).

    target_2.id : string
        The objectId of the second target object to which the first target
        object must be transferred.

    target_2.image : list of numpy arrays
        An image of the second target object to which the first target object
        must be transferred, given as a 3D RGB pixel array.

    target_2.info : list of strings
        Human-readable information describing the target object needed for the
        visualization interface.

    target_2.match_image : string
        Whether the image of the second target object (target_2.image) exactly
        matches the actual object in the scene. If false, then the actual
        second target object will be different in one way (for example, the
        image may depict a blue ball, but the actual object is a yellow ball,
        or a blue cube).
    """

    TRAVERSAL = "traversal"
    """
    NOT USED IN MCS EVAL 4+

    In a trial that has a traversal goal, you must find and move next to a
    target object. This may involve exploring the scene, and avoiding
    obstacles. These trials will demand a "common sense" understanding of
    self navigation (how to move and rotate yourself within a scene and around
    obstacles).

    Parameters
    ----------
    target.id : string
        The objectId of the target object to find and move next to.

    target.image : list of numpy arrays
        An image of the target object to find and move next to, given as a 3D
        RGB pixel array.

    target.info : list of strings
        Human-readable information describing the target object needed for the
        visualization interface.

    target.match_image : string
        Whether the image of the target object (target.image) exactly matches
        the actual target object in the scene. If false, then the actual object
        will be different in one way (for example, the image may depict a blue
        ball, but the actual object is a yellow ball, or a blue cube).
    """



from typing import Dict, List

from shapely import geometry

from .controller import DEFAULT_MOVE
from .goal_metadata import GoalCategory, GoalMetadata

GOAL_ACHIEVED = 1
GOAL_NOT_ACHIEVED = 0
STEP_PENALTY = 0.001


class Reward(object):
    '''Reward utility class'''
    @staticmethod
    def __get_object_from_list(objects: List[Dict],
                               target_id: str) -> Dict:
        '''
        Finds an object in a list. Uses a generator to return the first item
        or defaults to None if the target isn't found.

        Args:
            objects: list of object dictionaries
            target_id: str objectId of the object to find

        Returns:
            target: object dictionary if found or None
        '''
        return next((o for o in objects if o['objectId'] == target_id), None)

    @staticmethod
    def _convert_object_to_planar_polygon(
            goal_object: Dict) -> geometry.Polygon:
        '''
        Project goal object bounds (x,y,z) to an XZ planar polygon.

        Args:
            goal_object: dict

        Returns:
            polygons: shapely.geometry.Polygon
        '''
        bbox3d = goal_object['objectBounds']['objectBoundsCorners']
        # project to XZ plane
        xz_pts = [(pt['x'], pt['z']) for pt in bbox3d]
        polygon = geometry.MultiPoint(xz_pts).convex_hull

        return polygon

    @staticmethod
    def _calc_retrieval_reward(
            goal: GoalMetadata,
            objects: Dict,
            agent: Dict,
            performer_reach: float) -> int:
        '''
        Calculate the reward for the retrieval goal.

        The goal object must be in the agent's hand.

        Args:
            goal: GoalMetadata
            objects: Dict
            agent: Dict
            performer_reach: float

        Returns:
            int: 1 for goal achieved, 0 otherwise

        '''
        reward = GOAL_NOT_ACHIEVED
        goal_id = goal.metadata['target'].get('id', None)
        goal_object = Reward.__get_object_from_list(objects, goal_id)

        if goal_object and goal_object.get('isPickedUp', False):
            reward = GOAL_ACHIEVED
        return reward

    @staticmethod
    def _calc_traversal_reward(
            goal: GoalMetadata,
            objects: Dict,
            agent: Dict,
            performer_reach: float) -> int:
        '''
        Calculate the reward for the traversal goal.

        Agent must be within reach distance of an object edge to be
        considered near.

        Args:
            goal: GoalMetadata
            objects: Dict
            agent: Dict
            performer_reach: float

        Returns:
            int: 1 for goal achieved, 0 otherwise

        '''
        reward = GOAL_NOT_ACHIEVED
        goal_id = goal.metadata['target'].get('id', None)
        goal_object = Reward.__get_object_from_list(objects, goal_id)

        agent_pos = agent['position']
        agent_xz = geometry.Point(agent_pos['x'], agent_pos['z'])

        if goal_object is not None:
            goal_polygon = Reward._convert_object_to_planar_polygon(
                goal_object)
            polygonal_distance = agent_xz.distance(goal_polygon)
            reward = int(polygonal_distance <= performer_reach)

        return reward

    @staticmethod
    def _calc_transferral_reward(
            goal: GoalMetadata,
            objects: Dict,
            agent: Dict,
            performer_reach: float) -> int:
        '''
        Calculate the reward for the transferral goal.

        The action object must be next to or on top of the goal object. This
        depends on the relationship action verb.

        Args:
            goal: GoalMetadata
            objects: Dict
            agent: Dict
            performer_reach: float

        Returns:
            int: 1 for goal achieved, 0 otherwise

        '''
        reward = GOAL_NOT_ACHIEVED

        relationship = goal.metadata.get('relationship', None)
        if relationship is None or len(relationship) != 3:
            return GOAL_NOT_ACHIEVED

        # action object to goal object
        _, action, _ = relationship
        action_target = goal.metadata.get('target_1', None)
        action_id = action_target.get('id', None)
        goal_target = goal.metadata.get('target_2', None)
        goal_id = goal_target.get('id', None)
        action = action.lower()

        # objects = scene_metadata['objects']
        action_object = Reward.__get_object_from_list(objects, action_id)
        goal_object = Reward.__get_object_from_list(objects, goal_id)

        if goal_object is None or goal_object.get('isPickedUp', False):
            return GOAL_NOT_ACHIEVED

        if action_object is None or action_object.get('isPickedUp', False):
            return GOAL_NOT_ACHIEVED

        goal_polygon = Reward._convert_object_to_planar_polygon(
            goal_object)
        action_polygon = Reward._convert_object_to_planar_polygon(
            action_object)

        # actions are next_to or on_top_of (ie; action obj next to goal obj)
        if action == 'next to':
            polygonal_distance = action_polygon.distance(goal_polygon)
            reward = int(polygonal_distance <= DEFAULT_MOVE)
        elif action == 'on top of':
            # check that the action object center intersects the goal object
            # bounds and the y dimension of the target is above the goal
            action_obj_within_goal = action_polygon.intersects(goal_polygon)
            action_obj_above_goal = (action_object['position']['y'] >
                                     goal_object['position']['y'])
            if action_obj_within_goal and action_obj_above_goal:
                reward = GOAL_ACHIEVED

        return reward

    @staticmethod
    def _adjust_score_penalty(
            current_score: int,
            number_steps: int) -> float:
        '''
        Calculate the score penalty based on the number of steps,
        if the current step results in a reward being achieved do
        not penalize them for the step that resulted in the goal
        being achieved.

        Args:
            current_score: 1 or 0 depending if reward achieved
            number_steps: the current step count

        Returns:
            float: new score based off of step penalty

        '''
        if current_score == 1:
            return current_score - ((number_steps - 1) * STEP_PENALTY)
        else:
            return current_score - ((number_steps) * STEP_PENALTY)

    @staticmethod
    def _calculate_default_reward(
            goal: GoalMetadata,
            objects: Dict,
            agent: Dict,
            reach: float) -> int:
        '''Returns the default reward of 0; not achieved.'''
        return GOAL_NOT_ACHIEVED

    @staticmethod
    def calculate_reward(
            goal: GoalMetadata,
            objects: Dict,
            agent: Dict,
            number_steps: int,
            reach: float) -> float:
        '''
        Determine if the agent achieved the objective/task/goal.

        Args:
            goal: GoalMetadata
            objects: Dict
            agent: Dict
            reach: float

        Returns:
            int: reward is 1 if goal achieved, 0 otherwise

        '''

        category = None
        if goal is not None and goal.metadata:
            category = goal.metadata.get('category', None)

        switch = {
            GoalCategory.RETRIEVAL.value: Reward._calc_retrieval_reward,  # noqa: E501
            GoalCategory.TRANSFERRAL.value: Reward._calc_transferral_reward,  # noqa: E501
            GoalCategory.TRAVERSAL.value: Reward._calc_traversal_reward,  # noqa: E501
        }

        current_score = switch.get(category,
                                   Reward._calculate_default_reward)(goal,
                                                                     objects,
                                                                     agent,
                                                                     reach)
        return Reward._adjust_score_penalty(current_score, number_steps)



from .stringifier import Stringifier


class ObjectMetadata(object):
    """
    Defines metadata for an object in the MCS 3D environment.

    Attributes
    ----------
    uuid : string
        The unique ID of this object, used with some actions.
    color : dict
        The "r", "g", and "b" pixel values of this object in images from the
        StepMetadata's "object_mask_list".
    dimensions : list of dicts
        The dimensions of this object in the environment's 3D global
        coordinate system as a list of 8 points (dicts with "x", "y", and "z").
    direction : dict
        The direction vector of "x", "y", and "z" degrees between your position
        and this object's position (the difference in the two positions),
        normalized to 1. You can use the "x" and "y" as the "rotation" and
        "horizon" parameters (respectively) in a "RotateLook" action to face
        this object.
    distance : float
        DEPRECATED. Same as distance_in_steps. Please use distance_in_steps
        or distance_in_world.
    distance_in_steps : float
        The distance from you to this object in number of steps ("Move"
        actions) on the 2D X/Z movement grid.
    distance_in_world : float
        The distance from you to this object in the environment's 3D global
        coordinate system.
    held : boolean
        Whether you are holding this object.
    mass : float
        Haptic feedback. The mass of this object.
    material_list : list of strings
        Haptic feedback. The material(s) of this object.
        See :mod:`Material <machine_common_sense.Material>`.
    position : dict
        The "x", "y", and "z" coordinates for the global position of the
        center of this object's 3D model.
    rotation : dict
        This object's rotation angles around the "x", "y", and "z" axes
        in degrees.
    shape : string
        This object's shape in plain English.
    state_list : list of strings
        This object's state(s) from the current step in the scene. Sometimes
        used by objects with scripted behavior in passive scenes.
    texture_color_list : list of strings
        This object's colors, derived from its textures, in plain English.
    visible : boolean
        Whether you can see this object in your camera viewport.
    is_open : boolean
        Whether the object is open or not
    openable : boolean
        Whether the object can be opened
    """

    def __init__(
        self,
        uuid="",
        color=None,
        dimensions=None,
        direction=None,
        distance=-1.0,
        distance_in_steps=-1.0,
        distance_in_world=-1.0,
        held=False,
        mass=0.0,
        material_list=None,
        position=None,
        rotation=None,
        shape="",
        state_list=None,
        texture_color_list=None,
        visible=False,
        is_open=False,
        openable=False
    ):
        self.uuid = uuid
        self.color = {} if color is None else color
        self.dimensions = [] if dimensions is None else dimensions
        self.direction = {} if direction is None else direction
        self.distance = distance
        self.distance_in_steps = distance_in_steps
        self.distance_in_world = distance_in_world
        self.held = held
        self.mass = mass
        self.material_list = [] if material_list is None else material_list
        self.position = {} if position is None else position
        self.rotation = {} if rotation is None else rotation
        self.shape = shape
        self.state_list = [] if state_list is None else state_list
        self.texture_color_list = (
            [] if texture_color_list is None else texture_color_list
        )
        self.visible = visible
        self.is_open = is_open
        self.openable = openable

    def __str__(self):
        return Stringifier.class_to_str(self)

    # Allows converting the class to a dictionary, along with allowing
    #   certain fields to be left out of output file
    def __iter__(self):
        yield 'uuid', self.uuid
        yield 'color', self.color
        yield 'dimensions', self.dimensions
        yield 'direction', self.direction
        yield 'distance', self.distance
        yield 'distance_in_steps', self.distance_in_steps
        yield 'distance_in_world', self.distance_in_world
        yield 'held', self.held
        yield 'mass', self.mass
        yield 'material_list', self.material_list
        yield 'position', self.position
        yield 'rotation', self.rotation
        yield 'shape', self.shape
        yield 'state_list', self.state_list
        yield 'texture_color_list', self.texture_color_list
        yield 'visible', self.visible
        yield 'is_open', self.is_open
        yield 'openable', self.openable



import unittest

import ai2thor
import PIL

from machine_common_sense.config_manager import Vector3d
from machine_common_sense.plotter import TopDownPlotter, XZHeading


class TestTopDownPlotter(unittest.TestCase):

    def setUp(self):
        self.plotter = TopDownPlotter(
            team="test",
            scene_name="scene",
            room_size=Vector3d(x=10, y=4, z=10)
        )

    def test_convert_color_empty(self):
        color = self.plotter._convert_color('')
        self.assertEqual(color, "ivory")

    def test_convert_color_none(self):
        color = self.plotter._convert_color(None)
        self.assertEqual(color, "ivory")

    def test_convert_color_black(self):
        color = self.plotter._convert_color('black')
        self.assertEqual(color, "ivory")

    def test_convert_color(self):
        color = self.plotter._convert_color('red')
        self.assertEqual(color, "red")

    def test_plot_image_size(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img = self.plotter.plot(scene_event=scene_event, step_number=1)
        self.assertIsInstance(img, PIL.Image.Image)

    def test_plot_twice(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img1 = self.plotter.plot(scene_event=scene_event, step_number=1)
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 2.0,
                    'y': 0.0,
                    'z': 2.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }}
        scene_event = ai2thor.server.Event(metadata=metadata)
        img2 = self.plotter.plot(scene_event=scene_event, step_number=2)
        self.assertNotEqual(img1, img2)
        self.assertIsInstance(img1, PIL.Image.Image)
        self.assertIsInstance(img2, PIL.Image.Image)

    def test_calculate_heading_zero_degrees(self):
        heading = self.plotter._calculate_heading(
            0, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_ninety_degrees(self):
        heading = self.plotter._calculate_heading(
            90, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, -TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_one_eighty_degrees(self):
        heading = self.plotter._calculate_heading(
            180, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, -TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_two_seventy_degrees(self):
        heading = self.plotter._calculate_heading(
            270, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_three_sixty_degrees(self):
        heading = self.plotter._calculate_heading(
            360, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_seven_twenty_degrees(self):
        heading = self.plotter._calculate_heading(
            720, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_negative_ninety_degrees(self):
        heading = self.plotter._calculate_heading(
            -90, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.z, 0.0)
        self.assertIsInstance(heading, XZHeading)

    def test_calculate_heading_negative_seven_twenty_degrees(self):
        heading = self.plotter._calculate_heading(
            -720, TopDownPlotter.HEADING_LENGTH)
        self.assertAlmostEqual(heading.x, 0.0)
        self.assertAlmostEqual(heading.z, TopDownPlotter.HEADING_LENGTH)
        self.assertIsInstance(heading, XZHeading)

    def test_create_robot(self):
        robot_metadata = {
            'position': {
                'x': 2.70,
                'y': 1.23,
                'z': 3.14
            },
            'rotation': {
                'x': 0.0,
                'y': 78.0,
                'z': 0.0
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 2.70)
        self.assertAlmostEqual(robot.y, 1.23)
        self.assertAlmostEqual(robot.z, 3.14)
        self.assertAlmostEqual(robot.rotation, 78.0)

    def test_create_robot_missing_position(self):
        robot_metadata = {
            'rotation': {
                'x': 0.0,
                'y': 78.0,
                'z': 0.0
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 0.0)
        self.assertAlmostEqual(robot.y, 0.0)
        self.assertAlmostEqual(robot.z, 0.0)
        self.assertAlmostEqual(robot.rotation, 78.0)

    def test_create_robot_missing_rotation(self):
        robot_metadata = {
            'position': {
                'x': 2.70,
                'y': 1.23,
                'z': 3.14
            }
        }
        robot = self.plotter._create_robot(robot_metadata)

        self.assertAlmostEqual(robot.x, 2.70)
        self.assertAlmostEqual(robot.y, 1.23)
        self.assertAlmostEqual(robot.z, 3.14)
        self.assertAlmostEqual(robot.rotation, 0.0)

    def test_create_object(self):
        object_metadata = {
            'isPickedUp': True,
            'visibleInCamera': True,
            'objectId': 'test-uuid',
            'colorsFromMaterials': [
                'orange'
            ],
            'objectBounds': {
                'objectBoundsCorners': [
                    {'x': 0, 'y': 0, 'z': 0},
                    {'x': 0, 'y': 0, 'z': 1},
                    {'x': 1, 'y': 0, 'z': 1},
                    {'x': 1, 'y': 0, 'z': 0},
                    {'x': 0, 'y': 1, 'z': 0},
                    {'x': 0, 'y': 1, 'z': 1},
                    {'x': 1, 'y': 1, 'z': 1},
                    {'x': 1, 'y': 1, 'z': 0}
                ]
            }
        }
        obj = self.plotter._create_object(object_metadata)

        self.assertTrue(obj.held)
        self.assertTrue(obj.visible)
        self.assertEqual(obj.uuid, 'test-uuid')
        self.assertIsInstance(obj.bounds, list)
        self.assertEqual(len(obj.bounds), 8)
        self.assertEqual(obj.color, "orange")

    def test_create_object_empty(self):
        object_metadata = {}
        obj = self.plotter._create_object(object_metadata)

        self.assertIsNone(obj.held)
        self.assertIsNone(obj.visible)
        self.assertIsNone(obj.uuid)
        self.assertIsNone(obj.bounds)
        self.assertEqual(obj.color, "ivory")

    def test_find_plottable_objects_empty(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [],
            'structuralObjects': [],
            'agent': {
                'position': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                },
                'rotation': {
                    'x': 0.0,
                    'y': 0.0,
                    'z': 0.0
                }
            }
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 0)

    def test_find_plottable_objects_combined(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [
                {'objectId': 'test-uuid1'},
                {'objectId': 'test-uuid2'},
                {'objectId': 'test-uuid3'}
            ],
            'structuralObjects': [
                {'objectId': 'occluder1'},
                {'objectId': 'occluder2'},
                {'objectId': 'wall1'}
            ]
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 6)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('test-uuid')]), 3)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('occluder')]), 2)
        self.assertEqual(len(
            [k for k in filtered_objects
             if k['objectId'].startswith('wall')]), 1)

    def test_find_plottable_objects_filtered(self):
        metadata = {
            'screenWidth': 600,
            'screenHeight': 400,
            'objects': [
                {'objectId': 'test-uuid1'},
                {'objectId': 'test-uuid2'},
                {'objectId': 'test-uuid3'}
            ],
            'structuralObjects': [
                {'objectId': 'occluder1'},
                {'objectId': 'occluder2'},
                {'objectId': 'wall1'},
                {'objectId': 'future-object'},
                {'objectId': 'floor-gets-filtered'},
                {'objectId': 'ceiling-gets-filtered'}
            ]
        }
        scene_event = ai2thor.server.Event(metadata=metadata)
        filtered_objects = self.plotter._find_plottable_objects(scene_event)
        self.assertEqual(len(filtered_objects), 7)
        self.assertEqual(filtered_objects[0]['objectId'], 'occluder1')
        self.assertEqual(filtered_objects[1]['objectId'], 'occluder2')
        self.assertEqual(filtered_objects[2]['objectId'], 'wall1')
        self.assertEqual(filtered_objects[3]['objectId'], 'future-object')
        self.assertEqual(filtered_objects[4]['objectId'], 'test-uuid1')
        self.assertEqual(filtered_objects[5]['objectId'], 'test-uuid2')
        self.assertEqual(filtered_objects[6]['objectId'], 'test-uuid3')

    def test_scene_name(self):
        self.assertEqual(self.plotter._scene_name, "scene")

    def test_scene_name_prefix(self):
        plotter = TopDownPlotter(
            team="test",
            scene_name="prefix/scene",
            room_size=Vector3d(x=10, y=3, z=10)
        )

        self.assertEqual(plotter._scene_name, "scene")


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Tuple

from shapely import geometry

import machine_common_sense as mcs


class TestReward(unittest.TestCase):

    def test_default_reward(self):
        goal = mcs.GoalMetadata()
        reward = mcs.Reward.calculate_reward(
            goal, objects={}, agent={}, number_steps=1, reach=1.0)
        self.assertEqual(reward, -0.001)
        self.assertIsInstance(reward, float)

    def test_none_goal(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects={}, agent={}, number_steps=5, reach=1.0)
        self.assertEqual(reward, -0.005)
        self.assertIsInstance(reward, float)

    def test_penalty_step_calcuation(self):
        goal = None
        reward = mcs.Reward.calculate_reward(
            goal, objects={}, agent={}, number_steps=456, reach=1.0)
        penalty = 0 - ((456) * 0.001)
        self.assertEqual(reward, penalty)
        self.assertIsInstance(reward, float)

    def test_target_in_empty_object_list(self):
        obj_list = []
        target_id = ''
        self.assertIsNone(
            mcs.Reward._Reward__get_object_from_list(
                obj_list, target_id))

    def test_target_in_object_list(self):
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i)}
            obj_list.append(obj)
        target_id = '7'
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertTrue(results)
        self.assertIsInstance(results, dict)
        self.assertEqual(results['objectId'], target_id)

    def test_target_not_in_object_list(self):
        obj_list = []
        for i in range(10):
            obj_list.append({"objectId": str(i)})
        target_id = '111'  # not in list
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertIsNone(results)

    def test_duplicate_target_ids_in_list(self):
        obj_list = []
        for i in range(10):
            obj_list.append({"objectId": str(i)})
        # add duplicate object
        obj_list.append({"objectId": '7', "duplicate": True})
        target_id = '7'
        results = mcs.Reward._Reward__get_object_from_list(
            obj_list, target_id)
        self.assertTrue(results)
        self.assertIsInstance(results, dict)
        self.assertEqual(results['objectId'], target_id)
        # expect that method still returns the first instance and not the
        # duplicate
        self.assertFalse('duplicate' in results)

    def test_convert_object_to_planar_polygon(self):
        goal_object = {'objectBounds': {'objectBoundsCorners': []}}
        # create lower plane (y = 0)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})

        polygon = mcs.Reward._convert_object_to_planar_polygon(goal_object)

        self.assertIsInstance(polygon, geometry.Polygon)
        self.assertEqual(len(list(polygon.exterior.coords)), 5)
        self.assertIsInstance(list(polygon.exterior.coords)[0], Tuple)

    def test_convert_skewed_object_to_planar_polygon(self):
        goal_object = {'objectBounds': {'objectBoundsCorners': []}}

        # create lower plane (y = 0)
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 0.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 0.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1) but shifted over x and z by 1
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 2.0, 'y': 1.0, 'z': 1.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 2.0, 'y': 1.0, 'z': 2.0})
        goal_object['objectBounds']['objectBoundsCorners'].append(
            {'x': 1.0, 'y': 1.0, 'z': 2.0})

        polygon = mcs.Reward._convert_object_to_planar_polygon(goal_object)

        self.assertIsInstance(polygon, geometry.Polygon)
        self.assertEqual(len(list(polygon.exterior.coords)), 7)
        self.assertIsInstance(list(polygon.exterior.coords)[0], Tuple)

    def test_retrieval_reward(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i), 'isPickedUp': not i}
            obj_list.append(obj)
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_retrieval_reward_nothing_pickedup(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        for i in range(10):
            obj = {"objectId": str(i), 'isPickedUp': False}
            obj_list.append(obj)
        reward = mcs.Reward._calc_retrieval_reward(goal, obj_list, agent={},
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_traversal_reward(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []}}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 1.0})
            # create upper plane (y = 1)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 1.0})
            obj['position'] = {'x': 0.0 + i, 'z': 0.0}

            obj_list.append(obj)

        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_large_object_inside(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj = {"objectId": str(0), "objectBounds": {"objectBoundsCorners": []}}
        # create lower plane (y = 0)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 10.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 10.0})
        # create upper plane (y = 1)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 10.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 10.0})
        obj['position'] = {'x': 0.0, 'z': 0.0}
        obj_list = []
        obj_list.append(obj)

        agent = {'position': {'x': 5.0, 'y': 0.5, 'z': 5.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_large_object_long_side(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj = {"objectId": str(0), "objectBounds": {"objectBoundsCorners": []}}
        # create lower plane (y = 0)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})
        obj['position'] = {'x': 0.0, 'z': 0.0}
        obj_list = []
        obj_list.append(obj)

        agent = {'position': {'x': 10.1, 'y': 0.5, 'z': 1.1}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_large_object_long_side_out_of_reach(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj = {"objectId": str(0), "objectBounds": {"objectBoundsCorners": []}}
        # create lower plane (y = 0)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})
        obj['position'] = {'x': 0.0, 'z': 0.0}
        obj_list = []
        obj_list.append(obj)

        agent = {'position': {'x': 11.1, 'y': 0.5, 'z': 1.1}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_large_object_short_side(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj = {"objectId": str(0), "objectBounds": {"objectBoundsCorners": []}}
        # create lower plane (y = 0)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})
        obj['position'] = {'x': 0.0, 'z': 0.0}
        obj_list = []
        obj_list.append(obj)

        agent = {'position': {'x': -0.5, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_large_object_short_side_out_of_reach(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj = {"objectId": str(0), "objectBounds": {"objectBoundsCorners": []}}
        # create lower plane (y = 0)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 0.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 0.0, 'z': 1.0})
        # create upper plane (y = 1)
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 0.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 10.0, 'y': 1.0, 'z': 1.0})
        obj['objectBounds']['objectBoundsCorners'].append(
            {'x': 0.0, 'y': 1.0, 'z': 1.0})
        obj['position'] = {'x': 0.0, 'z': 0.0}
        obj_list = []
        obj_list.append(obj)

        agent = {'position': {'x': -1.5, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_outside_agent_reach(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '0'}
        obj_list = []
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': -1.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_traversal_reward_with_missing_target(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target'] = {'id': '111'}  # missing target
        obj_list = []
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_traversal_reward(goal, obj_list, agent,
                                                   performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_transferral_reward_with_missing_relationship(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target_1'] = {'id': '0'}
        goal.metadata['target_2'] = {'id': '1'}
        goal.metadata['relationship'] = []
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []}}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 1.0})
            # create upper plane (y = 1)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 1.0})
            obj['position'] = {'x': 0.5 + i, 'z': 0.5}
            obj_list.append(obj)
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_transferral_reward(goal, obj_list, agent,
                                                     performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_transferral_reward_next_to(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target_1'] = {'id': '0'}
        goal.metadata['target_2'] = {'id': '1'}
        goal.metadata['relationship'] = ['target_1', 'next to', 'target_2']
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []}}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 1.0})
            # create upper plane (y = 1)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 1.0})
            obj['position'] = {'x': 0.5 + i, 'z': 0.5}
            obj_list.append(obj)
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_transferral_reward(goal, obj_list, agent,
                                                     performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_transferral_reward_next_to_with_pickedup_object(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target_1'] = {'id': '0'}
        goal.metadata['target_2'] = {'id': '1'}
        goal.metadata['relationship'] = ['target_1', 'next to', 'target_2']
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []},
                "isPickedUp": True}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 0.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 0.0, 'z': 1.0})
            # create upper plane (y = 1)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0 + i, 'y': 1.0, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0 + i, 'y': 1.0, 'z': 1.0})
            obj['position'] = {'x': 0.5 + i, 'z': 0.5}
            obj_list.append(obj)
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_transferral_reward(goal, obj_list, agent,
                                                     performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)

    def test_transferral_reward_on_top_of(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target_1'] = {'id': '1'}
        goal.metadata['target_2'] = {'id': '0'}
        goal.metadata['relationship'] = ['target_1', 'on top of', 'target_2']
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []}}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 0.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 0.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 0.0 + i, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 0.0 + i, 'z': 1.0})
            # create upper plane (y = 1) + i
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 1.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 1.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 1.0 + i, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 1.0 + i, 'z': 1.0})
            obj['position'] = {'x': 0.5, 'y': 0.0 + i, 'z': 0.5}
            obj_list.append(obj)
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_transferral_reward(goal, obj_list, agent,
                                                     performer_reach=1.0)
        self.assertEqual(reward, 1)
        self.assertIsInstance(reward, int)

    def test_transferral_reward_on_top_of_with_pickedup_object(self):
        goal = mcs.GoalMetadata()
        goal.metadata['target_1'] = {'id': '0'}
        goal.metadata['target_2'] = {'id': '1'}
        goal.metadata['relationship'] = ['target_1', 'on top of', 'target_2']
        obj_list = []
        for i in range(10):
            obj = {
                "objectId": str(i),
                "objectBounds": {
                    "objectBoundsCorners": []},
                "isPickedUp": True}
            # create lower plane (y = 0)
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 0.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 0.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 0.0 + i, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 0.0 + i, 'z': 1.0})
            # create upper plane (y = 1) + i
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 1.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 1.0 + i, 'z': 0.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 1.0, 'y': 1.0 + i, 'z': 1.0})
            obj['objectBounds']['objectBoundsCorners'].append(
                {'x': 0.0, 'y': 1.0 + i, 'z': 1.0})
            obj['position'] = {'x': 0.5, 'y': 0.0 + i, 'z': 0.5}
            obj_list.append(obj)
        agent = {'position': {'x': -0.9, 'y': 0.5, 'z': 0.0}}
        reward = mcs.Reward._calc_transferral_reward(goal, obj_list, agent,
                                                     performer_reach=1.0)
        self.assertEqual(reward, 0)
        self.assertIsInstance(reward, int)


if __name__ == '__main__':
    unittest.main()



import unittest

import machine_common_sense as mcs


class TestValidation(unittest.TestCase):

    def test_is_in_range(self):
        self.assertEqual(mcs.Validation.is_in_range(0, 0, 1, 1234), 0)
        self.assertEqual(mcs.Validation.is_in_range(0.5, 0, 1, 1234), 0.5)
        self.assertEqual(mcs.Validation.is_in_range(1, 0, 1, 1234), 1)

        self.assertEqual(mcs.Validation.is_in_range(-1, 0, 1, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(1.01, 0, 1, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(100, 0, 1, 1234), 1234)

        self.assertEqual(mcs.Validation.is_in_range(2, 2, 4, 1234), 2)
        self.assertEqual(mcs.Validation.is_in_range(2.1, 2, 4, 1234), 2.1)
        self.assertEqual(mcs.Validation.is_in_range(4, 2, 4, 1234), 4)

        self.assertEqual(mcs.Validation.is_in_range(1.9, 2, 4, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(-3, 2, 4, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(4.1, 2, 4, 1234), 1234)

        self.assertEqual(mcs.Validation.is_in_range(-2, -2, 2, 1234), -2)
        self.assertEqual(mcs.Validation.is_in_range(0, -2, 2, 1234), 0)
        self.assertEqual(mcs.Validation.is_in_range(2, -2, 2, 1234), 2)

        self.assertEqual(mcs.Validation.is_in_range(-2.1, -2, 2, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(2.1, -2, 2, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(200, -2, 2, 1234), 1234)

        self.assertEqual(mcs.Validation.is_in_range(-4, -4, -2, 1234), -4)
        self.assertEqual(mcs.Validation.is_in_range(-2.1, -4, -2, 1234), -2.1)
        self.assertEqual(mcs.Validation.is_in_range(-2, -4, -2, 1234), -2)

        self.assertEqual(mcs.Validation.is_in_range(-5, -4, -2, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(-1, -4, -2, 1234), 1234)
        self.assertEqual(mcs.Validation.is_in_range(3, -4, -2, 1234), 1234)

    def test_is_number(self):
        self.assertEqual(mcs.Validation.is_number('0'), True)
        self.assertEqual(mcs.Validation.is_number('1'), True)
        self.assertEqual(mcs.Validation.is_number('12.34'), True)
        self.assertEqual(mcs.Validation.is_number('01'), True)
        self.assertEqual(mcs.Validation.is_number(''), False)
        self.assertEqual(mcs.Validation.is_number('asdf'), False)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

import machine_common_sense as mcs


class TestSerializer(unittest.TestCase):

    @staticmethod
    def _helper_get_step_metadata():
        with open('tests/test.msgpack', 'rb') as file:
            packed_bytes = file.read()

        serializer = mcs.SerializerMsgPack()
        return serializer.deserialize(packed_bytes)

    def test_serialization_msgpack(self):
        unpacked_metadata = TestSerializer._helper_get_step_metadata()
        assert abs(unpacked_metadata.reward - (-0.036000000000000004)
                   ) < 1e-04, 'Reward unexpected.'
        assert abs(unpacked_metadata.rotation -
                   0.0) < 1e-04, 'Rotation unexpected.'
        assert isinstance(unpacked_metadata.object_list[-1].shape, str
                          ), 'Shape type unexpected.'

    def test_serialization_json(self):
        unpacked_metadata = TestSerializer._helper_get_step_metadata()

        serializer = mcs.SerializerJson()
        json_dump = serializer.serialize(unpacked_metadata)
        unpacked_metadata = serializer.deserialize(json_dump)
        self.assertEqual(len(unpacked_metadata.depth_map_list), 1)
        self.assertIsInstance(unpacked_metadata.depth_map_list[0], np.ndarray)
        self.assertEqual(unpacked_metadata.depth_map_list[0].shape, (400, 600))
        self.assertIsInstance(unpacked_metadata, mcs.StepMetadata)
        self.assertAlmostEqual(unpacked_metadata.reward, -0.036000000000000004)
        self.assertAlmostEqual(unpacked_metadata.rotation, 0.0, delta=1e-04)
        self.assertIsInstance(unpacked_metadata.object_list[-1].shape, str)


if __name__ == '__main__':
    unittest.main()



import ast
import configparser  # noqa: F401
import logging
import os
from dataclasses import dataclass, field
from typing import List

import numpy as np
from marshmallow import Schema, fields, post_load

from .action import Action
from .goal_metadata import GoalMetadata

logger = logging.getLogger(__name__)


@dataclass
class Vector3d:
    # There is probably a class like this in python somewhere
    # but i don't know where it is.
    # TODO change later, potentially rename?
    x: float = 0
    y: float = 0
    z: float = 0


class ConfigManager(object):

    DEFAULT_CLIPPING_PLANE_FAR = 15.0
    DEFAULT_ROOM_DIMENSIONS = Vector3d(x=10, y=3, z=10)

    # Normal metadata plus metadata for all hidden objects
    CONFIG_METADATA_TIER_ORACLE = 'oracle'
    # No metadata, except for the images, depth masks, object masks,
    # and haptic/audio feedback
    CONFIG_METADATA_TIER_LEVEL_2 = 'level2'
    # No metadata, except for the images, depth masks, and haptic/audio
    # feedback
    CONFIG_METADATA_TIER_LEVEL_1 = 'level1'
    # No metadata, except for the images and haptic/audio
    # feedback
    CONFIG_METADATA_TIER_NONE = 'none'

    # Default metadata level if none specified, meant for use during
    # development
    CONFIG_METADATA_TIER_DEFAULT = 'default'

    CONFIG_FILE_ENV_VAR = 'MCS_CONFIG_FILE_PATH'
    METADATA_ENV_VAR = 'MCS_METADATA_LEVEL'
    DEFAULT_CONFIG_FILE = './mcs_config.ini'

    CONFIG_DEFAULT_SECTION = 'MCS'

    CONFIG_EVALUATION_NAME = 'evaluation_name'
    CONFIG_HISTORY_ENABLED = 'history_enabled'
    CONFIG_METADATA_TIER = 'metadata'
    CONFIG_NOISE_ENABLED = 'noise_enabled'
    CONFIG_SAVE_DEBUG_IMAGES = 'save_debug_images'
    CONFIG_SAVE_DEBUG_JSON = 'save_debug_json'
    CONFIG_SEED = 'seed'
    CONFIG_SIZE = 'size'
    CONFIG_TEAM = 'team'
    CONFIG_VIDEO_ENABLED = 'video_enabled'

    # Please keep the aspect ratio as 3:2 because the IntPhys scenes are built
    # on this assumption.
    SCREEN_WIDTH_DEFAULT = 600
    SCREEN_WIDTH_MIN = 450

    def __init__(self, config_file_path=None):
        # For config file, look for environment variable first,
        # then look for config_path parameter from constructor
        self._config_file = os.getenv(
            self.CONFIG_FILE_ENV_VAR, config_file_path)

        if(self._config_file is None):
            self._config_file = self.DEFAULT_CONFIG_FILE

        self._read_config_file()

        self._validate_screen_size()

    def _read_config_file(self):
        self._config = configparser.ConfigParser()
        if os.path.exists(self._config_file):
            self._config.read(self._config_file)
            logger.info('Config File Path: ' + self._config_file)
        else:
            logger.info('No Config File')

    def _validate_screen_size(self):
        if(self.get_size() < self.SCREEN_WIDTH_MIN):
            self._config.set(
                self.CONFIG_DEFAULT_SECTION,
                self.CONFIG_SIZE,
                str(self.SCREEN_WIDTH_DEFAULT)
            )

    def get_evaluation_name(self):
        return self._config.get(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_EVALUATION_NAME,
            fallback=''
        )

    def get_metadata_tier(self):
        # Environment variable override for metadata property
        metadata_env_var = os.getenv('MCS_METADATA_LEVEL', None)

        if(metadata_env_var is None):
            return self._config.get(
                self.CONFIG_DEFAULT_SECTION,
                self.CONFIG_METADATA_TIER,
                fallback='default'
            )

        return metadata_env_var

    def set_metadata_tier(self, mode):
        self._config.set(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_METADATA_TIER,
            mode
        )

    def get_seed(self):
        return self._config.getint(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SEED,
            fallback=None
        )

    def get_size(self):
        return self._config.getint(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SIZE,
            fallback=self.SCREEN_WIDTH_DEFAULT
        )

    def get_team(self):
        return self._config.get(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_TEAM,
            fallback=''
        )

    def is_history_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_HISTORY_ENABLED,
            fallback=True
        )

    def is_noise_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_NOISE_ENABLED,
            fallback=False
        )

    def is_save_debug_images(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SAVE_DEBUG_IMAGES,
            fallback=False
        )

    def is_save_debug_json(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_SAVE_DEBUG_JSON,
            fallback=False
        )

    def is_video_enabled(self):
        return self._config.getboolean(
            self.CONFIG_DEFAULT_SECTION,
            self.CONFIG_VIDEO_ENABLED,
            fallback=False
        )

    def is_depth_maps_enabled(self) -> bool:
        metadata_tier = self.get_metadata_tier()
        return metadata_tier in [
            self.CONFIG_METADATA_TIER_LEVEL_1,
            self.CONFIG_METADATA_TIER_LEVEL_2,
            self.CONFIG_METADATA_TIER_ORACLE,
        ]

    def is_object_masks_enabled(self) -> bool:
        metadata_tier = self.get_metadata_tier()
        return (
            metadata_tier != self.CONFIG_METADATA_TIER_LEVEL_1 and
            metadata_tier in
            [
                self.CONFIG_METADATA_TIER_LEVEL_2,
                self.CONFIG_METADATA_TIER_ORACLE,
            ]
        )

    def get_screen_width(self) -> int:
        return self.get_size()

    def get_screen_height(self) -> int:
        size = self.get_size()
        return int(size / 3 * 2)


class Vector3dSchema(Schema):
    x = fields.Float()
    y = fields.Float()
    z = fields.Float()

    @post_load
    def make_position_3d(self, data, **kwargs):
        return Vector3d(**data)


class RoomMaterialsSchema(Schema):
    front = fields.Str()
    left = fields.Str()
    right = fields.Str()
    back = fields.Str()

    @post_load
    def make_room_materials(self, data, **kwargs):
        return RoomMaterials(**data)


class PerformerStartSchema(Schema):
    position = fields.Nested(Vector3dSchema)
    rotation = fields.Nested(Vector3dSchema)

    @post_load
    def make_performer_start(self, data, **kwargs):
        return PerformerStart(**data)


class GoalObjectSchema(Schema):
    id = fields.Str()
    image = fields.Raw(allow_none=True)


class GoalMetadataSchema(Schema):
    choose = fields.List(fields.Str())
    target = fields.Nested(GoalObjectSchema)
    target_1 = fields.Nested(GoalObjectSchema)
    target_2 = fields.Nested(GoalObjectSchema)


class GoalSchema(Schema):
    action_list = fields.List(fields.Raw())
    answer = fields.Dict()  # UI property
    category = fields.Str()
    description = fields.Str()
    domainsInfo = fields.Dict()  # UI property
    habituation_total = fields.Int()
    info_list = fields.List(fields.Str())
    last_preview_phase_step = fields.Int()
    last_step = fields.Int()
    metadata = fields.Dict()
    objectsInfo = fields.Dict()  # UI property
    sceneInfo = fields.Dict()  # UI property
    skip_preview_phase = fields.Bool()
    task_list = fields.List(fields.Str())
    type_list = fields.List(fields.Str())

    @post_load
    def make_goal(self, data, **kwargs):
        return Goal(**data)


class ChangeMaterialConfigSchema(Schema):
    stepBegin = fields.Int()
    materials = fields.List(fields.Str())

    @post_load
    def make_change_material(self, data, **kwargs):
        return ChangeMaterialConfig(**data)


class ForceConfigSchema(Schema):
    stepBegin = fields.Int()
    stepEnd = fields.Int()
    vector = fields.Nested(Vector3dSchema)
    relative = fields.Bool()

    @post_load
    def make_force_config(self, data, **kwargs):
        return ForceConfig(**data)


class MoveConfigSchema(Schema):
    stepBegin = fields.Int()
    stepEnd = fields.Int()
    vector = fields.Nested(Vector3dSchema)

    @post_load
    def make_move_config(self, data, **kwargs):
        return MoveConfig(**data)


class OpenCloseConfigSchema(Schema):
    step = fields.Int()
    open = fields.Bool()

    @post_load
    def make_move_config(self, data, **kwargs):
        return OpenCloseConfig(**data)


class PhysicsConfigSchema(Schema):
    enable = fields.Bool()
    angularDrag = fields.Float()
    bounciness = fields.Float()
    drag = fields.Float()
    dynamicFriction = fields.Float()
    staticFriction = fields.Float()

    @post_load
    def make_physics_config(self, data, **kwargs):
        return PhysicsConfig(**data)


class FloorHolesAndTexturesXZConfigSchema(Schema):
    x = fields.Int()
    z = fields.Int()

    @post_load
    def make_holes_config(self, data, **kwargs):
        return FloorHolesAndTexturesXZConfig(**data)


class FloorTexturesConfigSchema(Schema):
    material = fields.Str()
    positions = fields.List(fields.Nested(FloorHolesAndTexturesXZConfigSchema))

    @post_load
    def make_floor_textures_config(self, data, **kwargs):
        return FloorTexturesConfig(**data)


class ShowConfigSchema(Schema):
    stepBegin = fields.Int()
    position = fields.Nested(Vector3dSchema)
    rotation = fields.Nested(Vector3dSchema)
    scale = fields.Nested(Vector3dSchema)
    boundingBox = fields.List(fields.Dict())  # debug property

    @post_load
    def make_show_config(self, data, **kwargs):
        return ShowConfig(**data)


class SizeConfigSchema(Schema):
    stepBegin = fields.Int()
    stepEnd = fields.Int()
    size = fields.Nested(Vector3dSchema)

    @post_load
    def make_size_config(self, data, **kwargs):
        return SizeConfig(**data)


class SingleStepConfigSchema(Schema):
    stepBegin = fields.Int()

    @post_load
    def make_single_step_config(self, data, **kwargs):
        return SingleStepConfig(**data)


class StepBeginEndConfigSchema(Schema):
    stepBegin = fields.Int()
    stepEnd = fields.Int()

    @post_load
    def make_step_begin_end_config(self, data, **kwargs):
        return StepBeginEndConfig(**data)


class TeleportConfigSchema(Schema):
    stepBegin = fields.Int()
    position = fields.Nested(Vector3dSchema)

    @post_load
    def make_teleport_config(self, data, **kwargs):
        return TeleportConfig(**data)


class TransformConfigSchema(Schema):
    position = fields.Nested(Vector3dSchema)
    rotation = fields.Nested(Vector3dSchema)
    scale = fields.Nested(Vector3dSchema)

    @post_load
    def make_transform_config(self, data, **kwargs):
        return TransformConfig(**data)


class SceneObjectSchema(Schema):
    id = fields.Str()
    type = fields.Str()  # should this be an enum?
    centerOfMass = fields.Nested(Vector3dSchema)
    changeMaterials = fields.List(fields.Nested(ChangeMaterialConfigSchema))
    debug = fields.Dict()
    forces = fields.List(fields.Nested(ForceConfigSchema))
    ghosts = fields.List(fields.Nested(StepBeginEndConfigSchema))
    hides = fields.List(fields.Nested(SingleStepConfigSchema))
    kinematic = fields.Bool()
    locationParent = fields.Str()
    mass = fields.Float()
    materials = fields.List(fields.Str())
    materialFile = fields.Str()  # deprecated; please use materials
    moveable = fields.Bool()
    moves = fields.List(fields.Nested(MoveConfigSchema))
    nullParent = fields.Nested(TransformConfigSchema)
    openable = fields.Bool()
    opened = fields.Bool()
    openClose = fields.List(fields.Nested(OpenCloseConfigSchema))
    physics = fields.Bool()
    physicsProperties = fields.Nested(PhysicsConfigSchema)
    pickupable = fields.Bool()
    receptacle = fields.Bool()
    resetCenterOfMass = fields.Bool()
    resizes = fields.List(fields.Nested(SizeConfigSchema))
    rotates = fields.List(fields.Nested(MoveConfigSchema))
    salientMaterials = fields.List(fields.Str())
    seesaw = fields.Bool()
    shows = fields.List(fields.Nested(ShowConfigSchema))
    shrouds = fields.List(fields.Nested(StepBeginEndConfigSchema))
    states = fields.List(fields.List(fields.Str(), allow_none=True))
    structure = fields.Bool()
    teleports = fields.List(fields.Nested(TeleportConfigSchema))
    togglePhysics = fields.List(fields.Nested(SingleStepConfigSchema))
    torques = fields.List(fields.Nested(MoveConfigSchema))

    # These are deprecated, but needed for Eval 3 backwards compatibility
    canContainTarget = fields.Bool()
    obstacle = fields.Bool()
    occluder = fields.Bool()
    positionY = fields.Float()
    stackTarget = fields.Bool()

    @post_load
    def make_scene_object(self, data, **kwargs):
        return SceneObject(**data)


class SceneConfigurationSchema(Schema):
    # The passive agent scenes can have a None ceilingMaterial
    ceilingMaterial = fields.Str(allow_none=True)
    debug = fields.Dict()
    floorMaterial = fields.Str()
    floorProperties = fields.Nested(PhysicsConfigSchema)
    goal = fields.Nested(GoalSchema)
    intuitivePhysics = fields.Bool()
    isometric = fields.Bool()
    name = fields.Str()
    objects = fields.List(fields.Nested(SceneObjectSchema))
    observation = fields.Bool()  # deprecated; please use intuitivePhysics
    performerStart = fields.Nested(PerformerStartSchema)
    roomDimensions = fields.Nested(Vector3dSchema)
    roomMaterials = fields.Nested(RoomMaterialsSchema)
    screenshot = fields.Bool()  # developer use only; for the image generator
    version = fields.Integer()
    wallMaterial = fields.Str()
    wallProperties = fields.Nested(PhysicsConfigSchema)
    holes = fields.List(fields.Nested(FloorHolesAndTexturesXZConfigSchema))
    floorTextures = fields.List(fields.Nested(FloorTexturesConfigSchema))

    # These are deprecated, but needed for Eval 3 backwards compatibility
    evaluation = fields.Str(allow_none=True)
    evaluationOnly = fields.Bool()
    evalName = fields.Str()
    hypercubeNumber = fields.Int()
    sceneNumber = fields.Int()
    sequenceNumber = fields.Int()
    training = fields.Bool()

    @post_load
    def make_scene_configuration(self, data, **kwargs):
        return SceneConfiguration(**data)


@dataclass
class RoomMaterials:
    front: str = None
    left: str = None
    right: str = None
    back: str = None


@dataclass
class PerformerStart:
    position: Vector3d
    rotation: Vector3d


@dataclass
class Goal:
    action_list: list = None
    answer: dict = None  # UI property
    category: str = None
    description: str = None
    domainsInfo: dict = None  # UI property
    habituation_total: int = None
    info_list: list = None
    last_preview_phase_step: int = None
    last_step: int = None
    # TODO metadata objects
    metadata: dict = None
    objectsInfo: dict = None  # UI property
    sceneInfo: dict = None  # UI property
    skip_preview_phase: bool = False
    task_list: List[str] = None
    type_list: List[str] = None


@dataclass
class ChangeMaterialConfig:
    stepBegin: int
    materials: List[str]


@dataclass
class ForceConfig:
    stepBegin: int
    stepEnd: int
    vector: Vector3d = Vector3d(0, 0, 0)
    relative: bool = False


@dataclass
class MoveConfig:
    stepBegin: int
    stepEnd: int
    vector: Vector3d = Vector3d(0, 0, 0)


@dataclass
class OpenCloseConfig:
    step: int
    open: bool


@dataclass
class PhysicsConfig:
    enable: bool = False
    angularDrag: float = None
    bounciness: float = None
    drag: float = None
    dynamicFriction: float = None
    staticFriction: float = None


@dataclass
class FloorHolesAndTexturesXZConfig:
    x: int = None
    z: int = None


@dataclass
class FloorTexturesConfig:
    material: str
    positions: List[FloorHolesAndTexturesXZConfig] = None


@dataclass
class ShowConfig:
    stepBegin: int
    position: Vector3d = Vector3d(0, 0, 0)
    rotation: Vector3d = Vector3d(0, 0, 0)
    scale: Vector3d = Vector3d(1, 1, 1)
    boundingBox: List[dict] = field(default_factory=list)  # debug property


@dataclass
class SizeConfig:
    stepBegin: int
    stepEnd: int
    size: Vector3d = Vector3d(1, 1, 1)


@dataclass
class SingleStepConfig:
    stepBegin: int


@dataclass
class StepBeginEndConfig:
    stepBegin: int
    stepEnd: int


@dataclass
class TeleportConfig:
    stepBegin: int
    position: Vector3d = Vector3d(0, 0, 0)


@dataclass
class TransformConfig:
    position: Vector3d = Vector3d(0, 0, 0)
    rotation: Vector3d = Vector3d(0, 0, 0)
    scale: Vector3d = Vector3d(1, 1, 1)


@dataclass
class SceneObject:
    id: str
    type: str  # should this be an enum?
    centerOfMass: Vector3d = None
    changeMaterials: List[ChangeMaterialConfig] = None
    debug: dict = None
    forces: List[ForceConfig] = None
    ghosts: List[StepBeginEndConfig] = None
    hides: List[SingleStepConfig] = None
    kinematic: bool = None
    locationParent: str = None
    mass: float = None
    materials: List[str] = None
    materialFile: str = None  # deprecated; please use materials
    # Docs say moveable's default is dependant on type.  That could
    # be a problem for the concrete classes.  Needs more review later
    moveable: bool = None
    moves: List[MoveConfig] = None
    nullParent: TransformConfig = None
    openable: bool = None
    opened: bool = None
    openClose: List[OpenCloseConfig] = None
    physics: bool = None
    physicsProperties: PhysicsConfig = None
    pickupable: bool = None
    receptacle: bool = None
    resetCenterOfMass: bool = None
    resizes: List[SizeConfig] = None
    rotates: List[MoveConfig] = None
    salientMaterials: List[str] = None
    seesaw: bool = None
    shows: List[ShowConfig] = None
    shrouds: List[StepBeginEndConfig] = None
    states: List[List[str]] = None
    structure: bool = None
    teleports: List[TeleportConfig] = None
    togglePhysics: List[SingleStepConfig] = None
    torques: List[MoveConfig] = None

    # These are deprecated, but needed for Eval 3 backwards compatibility
    canContainTarget: bool = None
    obstacle: bool = None
    occluder: bool = None
    positionY: float = None
    stackTarget: bool = None


@dataclass
class SceneConfiguration:
    '''Class for keeping track of scene configuration'''
    ceilingMaterial: str = None
    debug: dict = None
    floorMaterial: str = None
    floorProperties: PhysicsConfig = None
    goal: Goal = None  # TODO change to concrete class
    intuitivePhysics: bool = False
    isometric: bool = False
    name: str = None
    objects: List[SceneObject] = field(default_factory=list)
    observation: bool = False  # deprecated; please use intuitivePhysics
    performerStart: PerformerStart = None
    roomDimensions: Vector3d = field(
        default=ConfigManager.DEFAULT_ROOM_DIMENSIONS)
    roomMaterials: RoomMaterials = None
    screenshot: bool = False  # developer use only; for the image generator
    version: int = None
    wallMaterial: str = None
    wallProperties: PhysicsConfig = None
    holes: List[FloorHolesAndTexturesXZConfig] = field(default_factory=list)
    floorTextures: List[FloorTexturesConfig] = field(default_factory=list)

    # These are deprecated, but needed for Eval 3 backwards compatibility
    evaluation: str = None
    evaluationOnly: bool = None
    evalName: str = None
    hypercubeNumber: int = None
    sceneNumber: int = None
    sequenceNumber: int = None
    training: bool = None

    def retrieve_object_states(self,
                               object_id, step_number):
        """Return the state list at the current step for the object with the
        given ID from the scene configuration data, if any."""
        state_list_each_step = []
        # Retrieve the object's states from the scene configuration.
        for object_config in self.objects:
            if object_config.id == object_id:
                state_list_each_step = object_config.states or []
                break
        # Retrieve the object's states in the current step.
        if len(state_list_each_step) > step_number:
            state_list = state_list_each_step[step_number]
            # Validate the data type.
            if state_list is not None:
                if not isinstance(state_list, list):
                    return [state_list]
                return [str(state) for state in state_list]
        return []

    def retrieve_goal(self):
        if not self.goal:
            return self.update_goal_target_image(GoalMetadata())

        goal = self.goal

        # Transform action list data from strings to tuples.
        action_list = goal.action_list or []
        for index, action_list_at_step in enumerate(action_list):
            action_list[index] = [
                Action.input_to_action_and_params(action)
                if isinstance(action, str) else action
                for action in action_list_at_step
            ]

        if goal.category:
            # Backwards compatibility
            goal.metadata['category'] = goal.category

        return self.update_goal_target_image(
            GoalMetadata(
                action_list=action_list or None,
                category=goal.category or '',
                description=goal.description or '',
                habituation_total=goal.habituation_total or 0,
                last_preview_phase_step=(goal.last_preview_phase_step or 0),
                last_step=goal.last_step or None,
                metadata=goal.metadata or {},
            )
        )

    def update_goal_target_image(self, goal_output):
        target_name_list = ['target', 'target_1', 'target_2']

        for target_name in target_name_list:
            # need to convert goal image data from string to array
            if (
                target_name in goal_output.metadata and
                'image' in goal_output.metadata[target_name] and
                isinstance(goal_output.metadata[target_name]['image'], str)
            ):
                image_list_string = goal_output.metadata[target_name]['image']
                goal_output.metadata[target_name]['image'] = np.array(
                    ast.literal_eval(image_list_string)).tolist()

        return goal_output



import textwrap
import unittest

import machine_common_sense as mcs


class TestGoalMetadata(unittest.TestCase):

    str_output = '''    {
        "action_list": null,
        "category": "",
        "description": "",
        "habituation_total": 0,
        "last_preview_phase_step": 0,
        "last_step": null,
        "metadata": {}
    }'''

    @classmethod
    def setUpClass(cls):
        cls.goal_metadata = mcs.GoalMetadata()

    @classmethod
    def tearDownClass(cls):
        pass

    def test_action_list(self):
        self.assertIsNone(self.goal_metadata.action_list)

    def test_category(self):
        self.assertEqual(self.goal_metadata.category, '')

    def test_description(self):
        self.assertEqual(self.goal_metadata.description, '')

    def test_habituation_total(self):
        self.assertEqual(self.goal_metadata.habituation_total, 0)
        self.assertIsInstance(self.goal_metadata.habituation_total, int)

    def test_last_preview_phase(self):
        self.assertEqual(self.goal_metadata.last_preview_phase_step, 0)
        self.assertIsInstance(self.goal_metadata.last_preview_phase_step, int)

    def test_last_step(self):
        self.assertIsNone(self.goal_metadata.last_step)

    def test_metadata(self):
        self.assertFalse(self.goal_metadata.metadata)
        self.assertIsInstance(self.goal_metadata.metadata, dict)

    def test_str(self):
        self.assertEqual(str(self.goal_metadata),
                         textwrap.dedent(self.str_output))


if __name__ == '__main__':
    unittest.main()



import math
from dataclasses import dataclass
from typing import Dict, List

import ai2thor.server
import colour
import cv2
import numpy as np
import PIL.Image
import skimage.draw
from shapely import geometry

from machine_common_sense.config_manager import Vector3d


@dataclass
class XZHeading():
    x: float
    z: float


@dataclass
class Robot():
    x: float
    y: float
    z: float
    rotation: float


@dataclass
class Object():
    held: bool
    visible: bool
    uuid: str
    color: str
    bounds: list


class TopDownPlotter():

    DEFAULT_COLOR = colour.COLOR_NAME_TO_RGB['white']
    BACKGROUND_COLOR = colour.COLOR_NAME_TO_RGB['black']
    BORDER_COLOR = colour.COLOR_NAME_TO_RGB['white']
    CENTER_COLOR = colour.COLOR_NAME_TO_RGB['gray']
    GRID_COLOR = colour.COLOR_NAME_TO_RGB['darkslategray']
    ROBOT_COLOR = colour.COLOR_NAME_TO_RGB['red']
    GOAL_COLOR = colour.COLOR_NAME_TO_RGB['gold']

    ROBOT_PLOT_WIDTH = 0.2
    HEADING_LENGTH = 0.4
    PLOT_IMAGE_SIZE = 512

    FONT = cv2.FONT_HERSHEY_COMPLEX
    FONT_SCALE = 0.4
    FONT_THICKNESS = 1

    def __init__(self, team: str, scene_name: str, room_size: Vector3d):
        self._team = team
        if '/' in scene_name:
            scene_name = scene_name.rsplit('/', 1)[1]
        self._scene_name = scene_name
        self._room_size = room_size
        # create the room once as it is computationally expensive
        self.grid_img = self._initialize_plot()

    def plot(self, scene_event: ai2thor.server.Event,
             step_number: int,
             goal_id: str = None
             ) -> PIL.Image.Image:
        '''Create a plot of the room, objects and robot'''
        plt_img = self.grid_img.copy()
        plt_objects = self._find_plottable_objects(scene_event)
        plt_img = self._draw_objects(plt_img,
                                     plt_objects,
                                     goal_id)
        agent_metadata = scene_event.metadata.get('agent', None)
        plt_img = self._draw_robot(
            plt_img,
            agent_metadata)
        plt_img = self._draw_step_number(plt_img, step_number)
        return self._export_plot(img=plt_img)

    def _initialize_plot(self) -> np.ndarray:
        '''Create the initial plot with grid lines'''
        img = np.zeros(
            (self.PLOT_IMAGE_SIZE, self.PLOT_IMAGE_SIZE, 3),
            dtype=np.int8)
        return self._draw_grid_lines(img)

    def _draw_step_number(self, img: np.ndarray,
                          step_number: int) -> np.ndarray:
        '''Write the step number on the plot image'''
        (label_width, label_height), baseline = cv2.getTextSize(
            str(step_number), self.FONT, self.FONT_SCALE, self.FONT_THICKNESS)
        text_draw_buffer = 5
        cv2.putText(
            img=img,
            text=str(step_number),
            org=(img.shape[0] - label_width - text_draw_buffer,
                 label_height + baseline),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=self.FONT_SCALE,
            color=self.BORDER_COLOR,
            lineType=self.FONT_THICKNESS)
        return img

    def _draw_grid_lines(self, img: np.ndarray) -> np.ndarray:
        '''Draw room coordinate grid aligned with Unity units'''
        room_x, room_z = self._room_size.x, self._room_size.z
        # treating y image dimension as z since this is an XZ planar plot
        x_dim, z_dim, _ = img.shape
        self.center_x = int(x_dim / 2) - 1  # ex: 512 /2 = 256 - 1 = 255
        self.center_z = int(z_dim / 2) - 1  # ex: 256 /2 = 128 - 1 = 127
        largest_dimension = max(room_x, room_z)
        self.x_scale = (x_dim - 1) / largest_dimension
        self.z_scale = (z_dim - 1) / largest_dimension

        # outer buffer space to keep 1:1 aspect ratio
        buffer_x = (x_dim - (x_dim * room_x / largest_dimension)) / 2
        buffer_z = (z_dim - (z_dim * room_z / largest_dimension)) / 2

        img = self._draw_vertical_grid_lines(img, buffer_z)
        img = self._draw_horizontal_grid_lines(img, buffer_x)
        img = self._draw_room_border(img, buffer_x, buffer_z)
        return img

    def _draw_vertical_grid_lines(
            self, img: np.ndarray, buffer_z: int) -> np.ndarray:
        x_units = math.floor(self._room_size.x / 2)
        for x in range(x_units + 1):
            img_x = int(self.center_x + x * self.x_scale)
            rr, cc = skimage.draw.line(
                c0=img_x,
                r0=int(buffer_z),
                c1=img_x,
                r1=img.shape[0] - 1 - int(buffer_z))
            img[rr, cc] = self.CENTER_COLOR if x == 0 else self.GRID_COLOR

            img_x = int(self.center_x - x * self.x_scale)
            rr, cc = skimage.draw.line(
                c0=img_x,
                r0=int(buffer_z),
                c1=img_x,
                r1=img.shape[0] - 1 - int(buffer_z))
            img[rr, cc] = self.CENTER_COLOR if x == 0 else self.GRID_COLOR
        return img

    def _draw_horizontal_grid_lines(
            self, img: np.ndarray, buffer_x: int) -> np.ndarray:
        z_units = math.floor(self._room_size.z / 2)
        for z in range(z_units + 1):
            img_z = int(self.center_z + z * self.z_scale)
            rr, cc = skimage.draw.line(
                c0=int(buffer_x),
                r0=img_z,
                c1=img.shape[1] - 1 - int(buffer_x),
                r1=img_z)
            img[rr, cc] = self.CENTER_COLOR if z == 0 else self.GRID_COLOR

            img_z = int(self.center_z - z * self.z_scale)
            rr, cc = skimage.draw.line(
                c0=int(buffer_x),
                r0=img_z,
                c1=img.shape[1] - 1 - int(buffer_x),
                r1=img_z)
            img[rr, cc] = self.CENTER_COLOR if z == 0 else self.GRID_COLOR
        return img

    def _draw_room_border(self, img: np.ndarray,
                          buffer_x: int, buffer_z: int) -> np.ndarray:
        x_dim, z_dim, _ = img.shape
        rr, cc = skimage.draw.rectangle_perimeter(
            start=(buffer_z, buffer_x),
            end=(z_dim - buffer_z - 1, x_dim - buffer_x - 1),
            shape=img.shape[:2],
            clip=True)
        img[rr, cc] = self.BORDER_COLOR
        return img

    def _export_plot(self, img: np.ndarray) -> PIL.Image.Image:
        '''Export the plot to a PIL Image'''
        return PIL.Image.fromarray(img, "RGB")

    def _draw_robot(self, img: np.ndarray, robot_metadata: Dict) -> np.ndarray:
        '''Plot the robot position and heading'''
        if robot_metadata is None:
            return None
        robot = self._create_robot(robot_metadata)
        img = self._draw_robot_position(img, robot)
        img = self._draw_robot_heading(img, robot)
        return img

    def _draw_robot_position(self, img: np.ndarray,
                             robot: Robot) -> np.ndarray:
        '''Draw the robot's scene XZ position in the plot'''
        rr, cc = skimage.draw.disk(
            center=(
                self.center_z - robot.z * self.z_scale,
                self.center_x + robot.x * self.x_scale),
            radius=self.ROBOT_PLOT_WIDTH * self.x_scale,
            shape=img.shape[:2])
        img[rr, cc] = self.ROBOT_COLOR
        return img

    def _draw_robot_heading(self, img: np.ndarray, robot: Robot) -> np.ndarray:
        '''Draw the heading vector starting from the robot XZ position'''
        heading = self._calculate_heading(
            rotation_angle=360.0 - robot.rotation,
            heading_length=self.HEADING_LENGTH
        )
        z_point = int(self.center_z - robot.z * self.z_scale)
        x_point = int(self.center_x + robot.x * self.x_scale)
        rr, cc = skimage.draw.line(
            z_point,
            x_point,
            z_point - int(heading.z * self.x_scale),
            x_point + int(heading.x * self.x_scale))
        img[rr, cc] = self.ROBOT_COLOR
        return img

    def _find_plottable_objects(
            self, scene_event: ai2thor.server.Event) -> List:
        '''Find plottable objects from the scene data.

        Plottable objects include normal scene objects as well as
        occluder and wall structural objects.
        '''
        structural_objects = scene_event.metadata.get('structuralObjects', [])
        filtered_structural_objects = [
            obj for obj in structural_objects
            if not obj.get('objectId', '').startswith('ceiling') and not
            obj.get('objectId', '').startswith('floor')
        ]
        objects = scene_event.metadata.get('objects', [])
        return filtered_structural_objects + objects

    def _draw_objects(self, img: np.ndarray, objects: Dict,
                      goal_id: str = None) -> np.ndarray:
        '''Plot the object bounds for each object in the scene'''
        for o in objects:
            obj = self._create_object(o)
            if obj.bounds is not None:
                obj_pts = [(pt['x'], pt['z']) for pt in obj.bounds]
                polygon = geometry.MultiPoint(
                    obj_pts).convex_hull
                pts = polygon.exterior.coords
                img = self._draw_object_bounds(img, obj, pts)
                if goal_id is not None and o['objectId'] == goal_id:
                    img = self._draw_goal(img, pts)
        return img

    def _draw_goal(self, img: np.ndarray,
                   pts: List) -> np.ndarray:
        '''Draw the goal object of the scene'''
        cs, rs = map(list, zip(*pts))
        rr, cc = skimage.draw.polygon_perimeter(
            rs,
            cs,
            shape=img.shape[:2])
        img[rr, cc] = self.GOAL_COLOR
        return img

    def _draw_object_bounds(self, img: np.ndarray,
                            obj: Object, points: List) -> np.ndarray:
        '''Draw the scene object'''
        # convert list of tuples to list of rows and list of columns
        cs, rs = map(list, zip(*points))
        # convert room coordinates to image coordinates
        rs = list(map(lambda r: self.center_z - r * self.z_scale, rs))
        cs = list(map(lambda c: self.center_x + c * self.x_scale, cs))
        # use dictionary get method for color retrieval
        # if no match, then None will resort to default color
        clr = colour.COLOR_NAME_TO_RGB.get(obj.color.lower())

        # draw filled polygon if visible to the robot
        if(obj.visible):
            rr, cc = skimage.draw.polygon(
                rs,
                cs,
                shape=img.shape[:2])
        else:
            rr, cc = skimage.draw.polygon_perimeter(
                rs,
                cs,
                shape=img.shape[:2])
        img[rr, cc] = clr
        return img

    def _calculate_heading(self, rotation_angle: float,
                           heading_length: float) -> XZHeading:
        '''Calculate XZ heading vector from the rotation angle'''
        s = math.sin(math.radians(rotation_angle))
        c = math.cos(math.radians(rotation_angle))
        vec_x = 0 * c - heading_length * s
        vec_z = 0 * s + heading_length * c
        return XZHeading(vec_x, vec_z)

    def _create_robot(self, robot_metadata: Dict) -> Robot:
        '''Extract robot position and rotation information from the metadata'''
        position = robot_metadata.get('position')
        if position is not None:
            x = position.get('x')
            y = position.get('y')
            z = position.get('z')
        else:
            x = 0.0
            y = 0.0
            z = 0.0
        rotation = robot_metadata.get('rotation')
        rotation_y = rotation.get('y') if rotation is not None else 0.0
        return Robot(x, y, z, rotation_y)

    def _create_object(self, object_metadata: Dict) -> Object:
        '''Create the scene object from its metadata'''

        held = object_metadata.get('isPickedUp')
        visible = object_metadata.get('visibleInCamera')
        uuid = object_metadata.get('objectId')
        colors = object_metadata.get('colorsFromMaterials', [])
        if len(colors):
            color = self._convert_color(colors[0])
        else:
            color = self._convert_color('')
        bounds = object_metadata.get('objectBounds')
        corners = None if bounds is None else bounds.get('objectBoundsCorners')
        return Object(
            held=held,
            visible=visible,
            uuid=uuid,
            color=color,
            bounds=corners
        )

    def _convert_color(self, color: str) -> str:
        '''Convert color string to string'''
        if not color:
            color = 'ivory'

        if color == 'black':
            color = 'ivory'
        return color



# This is additional documentation for logging configuration dictionaries.
# Documentation on the format can be found here:
# https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
#
#
# Basic structure of logging.
#   Code should get and use a logger determined by the class name:
#     logger = logging.getLogger(__name__)
#   Loggers are configured here to pass log data zero or more handlers
#     Loggers have a level which filters out any logs below that level
#   Handlers determine where to send the data (console, files, sockets)
#     Handlers have a level which filters out any logs below that level
#     Handlers can reference a formatter
#   Formatters determine the format of the log message
#   Note: For a message to be logged, its level must be equal to or above
#     both the logger's level and the handler's level
#
# "root" is Essentially the default.  Anything that isn't defined elsewhere,
# will take these settings
#
# "loggers" Sets all machine_common_sense loggers to these settings and
# logs them into the listed handlers
#
# "handlers" take log messages and put them somewhere (console, files,
# sockets, syslog, smtp, http)
# See https://docs.python.org/3/library/logging.handlers.html for
# all the different built in handlers
#
# The file handlers have a nice feature to rotate log files so you
# don't use up all your disk space (maxBytes and backupCount).
#
# creating separate file handlers can be useful, but we don't seem to have
# enough differentiation between debug and info yet to make this useful
#
# Formatters just specify the format of the log message.  Sometimes its
# useful to have a lot of detail.  Sometimes it more annoying than useful.
# Simply change which formatter you want in the handlers
# section.

import ast
import logging
import os

logger = logging.getLogger(__name__)


class LoggingConfig():

    default_console_config = {
        "version": 1,
        "root": {
            "level": "DEBUG",
            "handlers": ["console"],
            "propagate": False
        },
        "loggers": {
            "machine_common_sense": {
                "level": "DEBUG",
                "handlers": ["console"],
                "propagate": False
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "brief",
                "level": "DEBUG",
                "stream": "ext://sys.stdout"
            }
        },
        "formatters": {
            "brief": {
                "format": "%(message)s"
            }
        }
    }

    @staticmethod
    def init_logging(log_config=None,
                     log_config_file="log.config.user.py"):
        """
        Initializes logging system.  Attempts to read file first,
        which should not be checked in and each user can have their own.
        If user file doesn't exist, then there is a base config that
        should be read.
        """
        init_message = ""
        if (os.path.exists(log_config_file)):
            with open(log_config_file, "r") as data:
                log_config = ast.literal_eval(data.read())
                init_message = "Loaded logging config from " + log_config_file
        elif log_config is not None:
            init_message = "Loaded provided logging config dictionary"
        if (log_config is None):
            log_config = LoggingConfig.get_default_console_config()
            init_message = "Loaded default logging config"

        if (not os.path.exists("logs")):
            os.mkdir("logs")
        logging.config.dictConfig(log_config)
        logger.info(init_message)

    @staticmethod
    def get_default_console_config():
        return LoggingConfig.default_console_config

    @staticmethod
    def get_errors_only_console_config():
        return {
            "version": 1,
            "root": {
                "level": "ERROR",
                "handlers": ["console"],
                "propagate": False
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "brief",
                    "level": "DEBUG",
                    "stream": "ext://sys.stdout"
                }
            },
            "formatters": {
                "brief": {
                    "format": "%(message)s"
                }
            }
        }

    @staticmethod
    def get_dev_logging_config():
        '''Note: This logging configuration needs the log directory to be
        created relative to the current working directory of the python
        execution.
        '''
        return {
            "version": 1,
            "root": {
                "level": "DEBUG",
                "handlers": ["console", "debug-file"],
                "propagate": False
            },
            "loggers": {
                "machine_common_sense": {
                    "level": "DEBUG",
                    "handlers": ["console", "debug-file"],
                    "propagate": False
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "brief",
                    "level": "DEBUG",
                    "stream": "ext://sys.stdout"
                },
                "debug-file": {
                    "level": "DEBUG",
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "precise",
                    "filename": "logs/mcs.debug.log",
                    "maxBytes": 10240000,
                    "backupCount": 3
                },

                "info-file": {
                    "level": "INFO",
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "precise",
                    "filename": "logs/mcs.info.log",
                    "maxBytes": 10240000,
                    "backupCount": 3
                }
            },
            "formatters": {
                "brief": {
                    "format": "%(message)s"
                },
                "precise": {
                    "format": "%(asctime)s <%(levelname)s>: %(message)s"
                },
                "full": {
                    "format": "[%(name)s] %(asctime)s <%(levelname)s>: " +
                    "%(message)s"
                }
            }
        }



import logging

logger = logging.getLogger(__name__)


class Validation():
    '''
    Support for tests and conditions on validation numbers and strings.
    '''
    @staticmethod
    def is_in_range(value, min_value, max_value, default_value, label=None):
        """
        Returns if the given value is within the given min and max; if not,
        returns the given default.

        Parameters
        ----------
        value : number
            The input value.
        min_value : number
            The min value.
        max_value : number
            The max value.
        default_value : number
            The default value.
        label : string
            A label for the input value.  If given, and if the input
            value is not within the range, will print an error.

        Returns
        -------
        number
        """
        if value > max_value or value < min_value:
            if label is not None:
                logger.debug(
                    f'Value of {label} needs to be between '
                    f'{min_value} and {max_value}. '
                    f'Current value {value} '
                    f'will be reset to {default_value}.')
            return default_value
        return value

    @staticmethod
    def is_number(value, label=None):
        """
        Returns if the given value is a number.

        Parameters
        ----------
        value :
            The input value.
        label : string
            A label for the input value.  If given, and if the
            input value is not a number, will print an error.

        Returns
        -------
        boolean
        """
        try:
            float(value)
            return True
        except ValueError:
            if label is not None:
                logger.debug(f'Value of {label}'
                             f' needs to be a number. Will be set to 0.')
            return False



import atexit
import datetime
import glob
import json
import logging
import os
import random
from typing import Dict, Union

import ai2thor.controller
import ai2thor.server
import marshmallow
import numpy as np

logger = logging.getLogger(__name__)


# How far the player can reach.  I think this value needs to be bigger
# than the MAX_MOVE_DISTANCE or else the player may not be able to move
# into a position to reach some objects (it may be mathematically impossible).
# TODO Reduce this number once the player can crouch down to reach and
# pickup small objects on the floor.

DEFAULT_MOVE = 0.1

from .action import Action
from .config_manager import (ConfigManager, SceneConfiguration,
                             SceneConfigurationSchema)
from .controller_events import (AfterStepPayload, BeforeStepPayload,
                                EndScenePayload, EventType, StartScenePayload)
from .controller_output_handler import ControllerOutputHandler
from .goal_metadata import GoalMetadata
from .step_metadata import StepMetadata
from .validation import Validation


def __reset_override(self, scene):
    # From https://github.com/allenai/ai2thor/blob/2.5.0/ai2thor/controller.py#L503-L525 # noqa: E501
    # Remove the error check: if scene not in self.scenes_in_build
    self.server.send(dict(action='Reset', sceneName=scene, sequenceId=0))
    self.last_event = self.server.receive()
    self.last_event = self.step(
        action='Initialize',
        **self.initialization_parameters
    )
    return self.last_event


ai2thor.controller.Controller.reset = __reset_override


def __image_depth_override(self, image_depth_data, **kwargs):
    # From https://github.com/NextCenturyCorporation/ai2thor/blob/47a9d0802861ba8d7a2a7a6d943a46db28ddbaab/ai2thor/server.py#L232-L240 # noqa: E501
    # The MCS depth shader in Unity is completely different now, so override
    # the original AI2-THOR depth image code. Just return what Unity sends us.
    return ai2thor.server.read_buffer_image(
        image_depth_data,
        self.screen_width,
        self.screen_height
    )


ai2thor.server.Event._image_depth = __image_depth_override


class NumpyAwareEncoderOverride(json.JSONEncoder):
    # From https://github.com/allenai/ai2thor/blob/bd35d2cb887faee8b87aa04bd9373b027eb39f17/ai2thor/server.py#L17-L24 # noqa: E501
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return np.asscalar(obj)
        return super(NumpyAwareEncoderOverride, self).default(obj)


ai2thor.server.NumpyAwareEncoder = NumpyAwareEncoderOverride


class Controller():
    """
    MCS Controller class implementation for the MCS wrapper of the AI2-THOR
    library.

    https://ai2thor.allenai.org/ithor/documentation/

    Parameters
    ----------
    config_file_path: str, optional
        Path to configuration file to read in and set various properties,
        such as metadata level and whether or not to save history files
        (default None)
    """

    # AI2-THOR creates a square grid across the scene that is
    # uses for "snap-to-grid" movement. (This value may not
    # really matter because we set continuous to True in
    # the step input.)
    GRID_SIZE = 0.1

    MAX_FORCE = 50.0

    DEFAULT_HORIZON = 0
    DEFAULT_ROTATION = 0
    DEFAULT_FORCE = 0.5
    DEFAULT_AMOUNT = 0.5
    DEFAULT_IMG_COORD = 0
    DEFAULT_OBJECT_MOVE_AMOUNT = 1

    MAX_FORCE = 1
    MIN_FORCE = 0
    MAX_AMOUNT = 1
    MIN_AMOUNT = 0

    ROTATION_KEY = 'rotation'
    HORIZON_KEY = 'horizon'
    FORCE_KEY = 'force'
    AMOUNT_KEY = 'amount'

    OBJECT_IMAGE_COORDS_X_KEY = 'objectImageCoordsX'
    OBJECT_IMAGE_COORDS_Y_KEY = 'objectImageCoordsY'
    RECEPTACLE_IMAGE_COORDS_X_KEY = 'receptacleObjectImageCoordsX'
    RECEPTACLE_IMAGE_COORDS_Y_KEY = 'receptacleObjectImageCoordsY'

    # used for EndHabituation teleport
    TELEPORT_X_POS = 'xPosition'
    TELEPORT_Z_POS = 'zPosition'
    TELEPORT_Y_ROT = 'yRotation'

    # Hard coding actions that effect MoveMagnitude so the appropriate
    # value is set based off of the action
    # TODO: Move this to an enum or some place, so that you can determine
    # special move interactions that way
    FORCE_ACTIONS = ["ThrowObject", "PushObject", "PullObject"]
    OBJECT_MOVE_ACTIONS = ["CloseObject", "OpenObject"]
    MOVE_ACTIONS = ["MoveAhead", "MoveLeft", "MoveRight", "MoveBack"]

    def __init__(self, unity_app_file_path, config):

        self._subscribers = []

        self._end_scene_not_registered = True

        self._controller = ai2thor.controller.Controller(
            quality='Medium',
            fullscreen=False,
            # The headless flag does not work for me
            headless=False,
            local_executable_path=unity_app_file_path,
            width=config.get_screen_width(),
            height=config.get_screen_height(),
            # Set the name of our Scene in our Unity app
            scene='MCS',
            logs=True,
            # This constructor always initializes a scene, so add a scene
            # config to ensure it doesn't error
            sceneConfig={
                "objects": []
            }
        )

        if not self._controller:
            raise Exception('AI2-THOR/Unity Controller failed to initialize')

        self._on_init()
        self._set_config(config)

    def subscribe(self, subscriber):
        if subscriber not in self._subscribers:
            self._subscribers.append(subscriber)

    def remove_all_event_handlers(self):
        self._subscribers = []

    def _publish_event(self, event_type: EventType,
                       payload: Union[StartScenePayload, BeforeStepPayload,
                                      AfterStepPayload,
                                      EndScenePayload]):
        for subscriber in self._subscribers:
            # TODO should we make a deep copy of the payload so subscribers
            # cannot change source data?
            # seems like performance vs safety

            try:
                subscriber.on_event(event_type, payload)
            except Exception as msg:
                logger.error(
                    f"Error in event with type={event_type}" +
                    f" to subscriber={type(subscriber)}",
                    exc_info=msg)

    def _create_event_payload_kwargs(self) -> dict:
        return {"step_number": self.__step_number,
                "config": self._config,
                "scene_config": self._scene_config}

    def _create_post_step_event_payload_kwargs(
            self, wrapped_step, step_metadata, step_output: StepMetadata,
            restricted_step_output: StepMetadata) -> dict:
        args = self._create_event_payload_kwargs()
        args['output_folder'] = self.__output_folder
        args['timestamp'] = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        args['wrapped_step'] = wrapped_step
        args['step_metadata'] = step_metadata
        args['step_output'] = step_output
        args['restricted_step_output'] = restricted_step_output
        args['goal'] = self._goal
        return args

    # Pixel coordinates are expected to start at the top left, but
    # in Unity, (0,0) is the bottom left.

    def _convert_y_image_coord_for_unity(self, y_coord):
        if(y_coord != 0):
            return self._config.get_screen_height() - y_coord
        else:
            return y_coord

    def _on_init(self):
        self._goal = GoalMetadata()
        self.__habituation_trial = 1
        # Output folder used to save debug image, video, and JSON files.
        self.__output_folder = None
        self._scene_config = None
        self.__step_number = 0

    def _set_config(self, config):
        '''Allows config to be changed without changing the controller and
        attached Unity process.  This typically should only be called by the
        MCS package itself.

        For users, call machine_common_sense.change_config()
        '''
        self._config = config

        self._output_handler = ControllerOutputHandler(self._config)

        self.__noise_enabled = self._config.is_noise_enabled()
        self.__seed = self._config.get_seed()

        if self.__seed:
            random.seed(self.__seed)

    def end_scene(
        self,
        rating: Union[float, int, str] = None,
        score: float = 1.0,
        report: Dict[int, object] = None
    ):
        """
        Ends the current scene.  Calling end_scene() before calling
        start_scene() will do nothing.

        Parameters
        ----------
        rating : float or int or string, required
            The plausibility rating to classify a passive / VoE scene as either
            plausible or implausible. Not used for any interactive scenes. For
            passive agent scenes, this rating should be continuous, from 0.0
            (completely implausible) to 1.0 (completely plausible). For other
            passive scenes, this rating must be binary, either 0 (implausible)
            or 1 (plausible). Please note that end-of-scene ratings are
            required for all passive / VoE scenes. (default None)
        score : float, optional
            The continuous plausibility score between 0.0 (completely
            implausible) and 1.0 (completely plausible). End-of-scene scores
            are required for all passive / VoE scenes except agent scenes.
            Not used for any interactive scenes or passive agent scenes.
            (default 1.0)

            Note: when an issue causes the program to exit prematurely or
            end_scene isn't properly called but history_enabled is true,
            this value will be written to file as -1.
        report : Dict[int, object], optional
            Variable for retrospective per frame reporting for passive / VoE
            scenes. Not used for any interactive scenes or passive agent
            scenes. (default None)

            Key is an int representing a step/frame number from output step
            metadata, starting at 1. Value or payload contains:

                * rating : float or int or string, optional
                    The plausibility rating to classify a passive / VoE scene
                    as either plausible or implausible. Not used for any
                    interactive scenes. For passive agent scenes, this rating
                    should be continuous, from 0.0 (completely implausible) to
                    1.0 (completely plausible). For other passive scenes, this
                    rating must be binary, either 0 (implausible) or 1
                    (plausible). Please note that frame-by-frame ratings are no
                    longer required for any scenes (but end-of-scene ratings
                    are). (default None)
                * score : float, optional
                    The continuous plausibility score between 0.0 (completely
                    implausible) and 1.0 (completely plausible). Frame-by-frame
                    scores are required for all passive / VoE scenes except
                    agent scenes. Not used for any interactive scenes or
                    passive agent scenes. (default None)
                * violations_xy_list : List[Dict[str, float]], optional
                    A list of one or more (x, y) locations
                    (ex: [{"x": 1, "y": 3.4}]), each representing a potential
                    violation-of-expectation. These locations are required for
                    all passive / VoE scenes except agent scenes. Not used for
                    any interactive scenes or passive agent scenes.
                    (default None)
                * internal_state : object, optional
                    A properly formatted json object representing various kinds
                    of internal states at a particular moment. Examples include
                    the estimated position of the agent, current map of the
                    world, etc. (default None)

            Example report:

            {

                    1: {

                        "rating": 1,

                        "score": 0.75,

                        "violations_xy_list": [{"x": 1,"y": 1}],

                        "internal_state": {"test": "some state"}

                    }

            }

        """
        payloadArgs = self._create_event_payload_kwargs()
        payloadArgs['rating'] = str(rating)
        payloadArgs['score'] = score
        payloadArgs['report'] = report

        self._publish_event(
            EventType.ON_END_SCENE,
            EndScenePayload(
                **payloadArgs))

        if (not self._end_scene_not_registered):
            atexit.unregister(self.end_scene)
            self._end_scene_not_registered = True

    def _convert_scene_config(self, config_data) -> SceneConfiguration:
        if isinstance(config_data, SceneConfiguration):
            return config_data
        schema = SceneConfigurationSchema()
        return schema.load(config_data)

    def start_scene(self, config_data):
        """
        Starts a new scene using the given scene configuration data dict and
        returns the scene output data object.

        Parameters
        ----------
        config_data : SceneConfiguration or dict that can be serialized to
            SceneConfiguration
            The MCS scene configuration data for the scene to start.

        Returns
        -------
        StepMetadata
            The output data object from the start of the scene (the output from
            an "Initialize" action).
        """

        scene_config = self._convert_scene_config(config_data)

        self._scene_config = scene_config
        self.__habituation_trial = 1
        self.__step_number = 0
        self._goal = self._scene_config.retrieve_goal()

        skip_preview_phase = (scene_config.goal is not None and
                              scene_config.goal.skip_preview_phase)

        if (self.isFileWritingEnabled()):
            os.makedirs('./' + scene_config.name, exist_ok=True)
            self.__output_folder = './' + scene_config.name + '/'
            file_list = glob.glob(self.__output_folder + '*')
            for file_path in file_list:
                os.remove(file_path)

        sc = SceneConfigurationSchema(
            unknown=marshmallow.EXCLUDE).dump(
            scene_config)
        sc = self._remove_none(sc)
        wrapped_step = self.wrap_step(
            action='Initialize',
            sceneConfig=sc
        )
        step_output = self._controller.step(wrapped_step)

        self._output_handler.set_scene_config(scene_config)
        (pre_restrict_output, output) = self._output_handler.handle_output(
            step_output, self._goal, self.__step_number,
            self.__habituation_trial)

        if not skip_preview_phase:
            if (self._goal is not None and
                    self._goal.last_preview_phase_step > 0):
                image_list = output.image_list
                depth_map_list = output.depth_map_list
                object_mask_list = output.object_mask_list

                logger.debug('STARTING PREVIEW PHASE...')

                for _ in range(self._goal.last_preview_phase_step):
                    output = self.step('Pass')
                    image_list = image_list + output.image_list
                    depth_map_list = depth_map_list + output.depth_map_list
                    object_mask_list = (object_mask_list +
                                        output.object_mask_list)

                logger.debug('ENDING PREVIEW PHASE')

                output.image_list = image_list
                output.depth_map_list = depth_map_list
                output.object_mask_list = object_mask_list

            logger.debug('NO PREVIEW PHASE')

            # TODO Should this be in the if block?  Now that we are using
            # subscribers, we may want to always register
            if(self._end_scene_not_registered is True and
                    self._config.is_history_enabled()):
                # make sure history file is written when program exits
                atexit.register(self.end_scene, rating="", score=-1)
                self._end_scene_not_registered = False

        payloadArgs = self._create_post_step_event_payload_kwargs(
            wrapped_step, step_output, pre_restrict_output, output)
        payload = StartScenePayload(**payloadArgs)
        self._publish_event(
            EventType.ON_START_SCENE, payload)

        return output

    def isFileWritingEnabled(self):
        return self._scene_config.name is not None and (
            self._config.is_save_debug_images() or
            self._config.is_save_debug_json() or
            self._config.is_video_enabled()
        )

    # TODO: may need to reevaluate validation strategy/error handling in the
    # future
    """
    Need a validation/conversion step for what ai2thor will accept as input
    to keep parameters more simple for the user (in this case, wrapping
    rotation degrees into an object)
    """

    def validate_and_convert_params(self, action, **kwargs):
        moveMagnitude = DEFAULT_MOVE
        rotation = kwargs.get(self.ROTATION_KEY, self.DEFAULT_ROTATION)
        horizon = kwargs.get(self.HORIZON_KEY, self.DEFAULT_HORIZON)
        amount = kwargs.get(
            self.AMOUNT_KEY,
            self.DEFAULT_OBJECT_MOVE_AMOUNT
            if action in self.OBJECT_MOVE_ACTIONS
            else self.DEFAULT_AMOUNT
        )
        force = kwargs.get(self.FORCE_KEY, self.DEFAULT_FORCE)

        objectImageCoordsX = kwargs.get(
            self.OBJECT_IMAGE_COORDS_X_KEY, self.DEFAULT_IMG_COORD)
        objectImageCoordsY = kwargs.get(
            self.OBJECT_IMAGE_COORDS_Y_KEY, self.DEFAULT_IMG_COORD)
        receptacleObjectImageCoordsX = kwargs.get(
            self.RECEPTACLE_IMAGE_COORDS_X_KEY, self.DEFAULT_IMG_COORD)
        receptacleObjectImageCoordsY = kwargs.get(
            self.RECEPTACLE_IMAGE_COORDS_Y_KEY, self.DEFAULT_IMG_COORD)

        if not Validation.is_number(amount, self.AMOUNT_KEY):
            # The default for open/close is 1, the default for "Move" actions
            # is 0.5
            if action in self.OBJECT_MOVE_ACTIONS:
                amount = self.DEFAULT_OBJECT_MOVE_AMOUNT
            else:
                amount = self.DEFAULT_AMOUNT

        if not Validation.is_number(force, self.FORCE_KEY):
            force = self.DEFAULT_FORCE

        # Check object directions are numbers
        if not Validation.is_number(
                objectImageCoordsX,
                self.OBJECT_IMAGE_COORDS_X_KEY):
            objectImageCoordsX = self.DEFAULT_IMG_COORD

        if not Validation.is_number(
                objectImageCoordsY,
                self.OBJECT_IMAGE_COORDS_Y_KEY):
            objectImageCoordsY = self.DEFAULT_IMG_COORD

        # Check receptacle directions are numbers
        if not Validation.is_number(
                receptacleObjectImageCoordsX,
                self.RECEPTACLE_IMAGE_COORDS_X_KEY):
            receptacleObjectImageCoordsX = self.DEFAULT_IMG_COORD

        if not Validation.is_number(
                receptacleObjectImageCoordsY,
                self.RECEPTACLE_IMAGE_COORDS_Y_KEY):
            receptacleObjectImageCoordsY = self.DEFAULT_IMG_COORD

        amount = Validation.is_in_range(
            amount,
            self.MIN_AMOUNT,
            self.MAX_AMOUNT,
            self.DEFAULT_AMOUNT,
            self.AMOUNT_KEY)
        force = Validation.is_in_range(
            force,
            self.MIN_FORCE,
            self.MAX_FORCE,
            self.DEFAULT_FORCE,
            self.FORCE_KEY)

        # TODO Consider the current "head tilt" value while validating the
        # input "horizon" value.

        # Set the Move Magnitude to the appropriate amount based on the action
        if action in self.FORCE_ACTIONS:
            moveMagnitude = force * self.MAX_FORCE

        if action in self.OBJECT_MOVE_ACTIONS:
            moveMagnitude = amount

        if action in self.MOVE_ACTIONS:
            moveMagnitude = DEFAULT_MOVE

        # Add in noise if noise is enable
        if self.__noise_enabled:
            rotation = rotation * (1 + self.generate_noise())
            horizon = horizon * (1 + self.generate_noise())
            moveMagnitude = moveMagnitude * (1 + self.generate_noise())

        rotation_vector = {'y': rotation}
        object_vector = {
            'x': float(objectImageCoordsX),
            'y': self._convert_y_image_coord_for_unity(
                float(objectImageCoordsY)),
        }

        receptacle_vector = {
            'x': float(receptacleObjectImageCoordsX),
            'y': self._convert_y_image_coord_for_unity(
                float(receptacleObjectImageCoordsY)
            ),
        }

        teleportRotInput = kwargs.get(self.TELEPORT_Y_ROT)
        teleportPosXInput = kwargs.get(self.TELEPORT_X_POS)
        teleportPosZInput = kwargs.get(self.TELEPORT_Z_POS)

        teleportRotation = None
        teleportPosition = None

        if teleportRotInput is not None and Validation.is_number(
                teleportRotInput):
            teleportRotation = {'y': kwargs.get(self.TELEPORT_Y_ROT)}
        if (teleportPosXInput is not None and
                Validation.is_number(teleportPosXInput) and
                teleportPosZInput is not None and
                Validation.is_number(teleportPosZInput)):
            teleportPosition = {'x': teleportPosXInput, 'z': teleportPosZInput}
        return dict(
            objectId=kwargs.get("objectId", None),
            receptacleObjectId=kwargs.get("receptacleObjectId", None),
            rotation=rotation_vector,
            horizon=horizon,
            teleportRotation=teleportRotation,
            teleportPosition=teleportPosition,
            moveMagnitude=moveMagnitude,
            objectImageCoords=object_vector,
            receptacleObjectImageCoords=receptacle_vector
        )

    # Override
    def step(self, action: str, **kwargs) -> StepMetadata:
        """
        Runs the given action within the current scene.

        Parameters
        ----------
        action : string
            A selected action string from the list of available actions.
        **kwargs
            Zero or more key-and-value parameters for the action.

        Returns
        -------
        StepMetadata
            The MCS output data object from after the selected action and the
            physics simulation were run. Returns None if you have passed the
            "last_step" of this scene.
        """
        if (self._goal.last_step is not None and
                self._goal.last_step == self.__step_number):
            logger.warning(
                "You have passed the last step for this scene. "
                "Ignoring your action. Please call controller.end_scene() "
                "now.")
            return None

        if ',' in action:
            action, kwargs = Action.input_to_action_and_params(action)

        action_list = self._goal.retrieve_action_list_at_step(
            self.__step_number)
        # Only continue with this action step if the given action and
        # parameters are in the restricted action list.

        continue_with_step = False
        for restricted_action, restricted_params in action_list:
            if action == restricted_action:
                # If this action doesn't have restricted parameters, or each
                # input parameter is in the restricted parameters, it's OK.
                if len(restricted_params.items()) == 0 or all([
                    restricted_params.get(key) == value
                    for key, value in kwargs.items()
                ]):
                    continue_with_step = True
                    break
        if not continue_with_step:
            logger.warning(
                f"The given action '{action}' with parameters "
                f"'{kwargs}' isn't in the action_list. Ignoring your action. "
                f"Please call controller.step() with an action in the "
                f"action_list. Possible actions at step {self.__step_number}:"
            )
            for action_data in action_list:
                logger.warning(f'    {action_data}')
            return None

        self.__step_number += 1

        payloadArgs = self._create_event_payload_kwargs()
        payloadArgs['action'] = action
        payloadArgs['habituation_trial'] = self.__habituation_trial
        payloadArgs['goal'] = self._goal
        self._publish_event(
            EventType.ON_BEFORE_STEP,
            BeforeStepPayload(**payloadArgs))

        params = self.validate_and_convert_params(action, **kwargs)

        # Only call mcs_action_to_ai2thor_action AFTER calling
        # validate_and_convert_params
        action = self.mcs_action_to_ai2thor_action(action)

        if (action == 'EndHabituation'):
            self.__habituation_trial += 1

        if (self._goal.last_step is not None and
                self._goal.last_step == self.__step_number):
            logger.warning(
                "This is your last step for this scene. All "
                "your future actions will be skipped. Please call "
                "controller.end_scene() now.")

        step_action = self.wrap_step(action=action, **params)
        step_output = self._controller.step(step_action)

        (pre_restrict_output, output) = self._output_handler.handle_output(
            step_output, self._goal, self.__step_number,
            self.__habituation_trial)

        payloadArgs = self._create_post_step_event_payload_kwargs(
            step_action, step_output, pre_restrict_output, output)
        payloadArgs['ai2thor_action'] = action
        payloadArgs['step_params'] = params
        payloadArgs['action_kwargs'] = kwargs
        self._publish_event(
            EventType.ON_AFTER_STEP,
            AfterStepPayload(**payloadArgs))

        return output

    def mcs_action_to_ai2thor_action(self, action):
        if action == Action.CLOSE_OBJECT.value:
            # The AI2-THOR Python library has buggy error checking
            # specifically for the CloseObject action,
            # so just use our own custom action here.
            return "MCSCloseObject"

        if action == Action.DROP_OBJECT.value:
            return "DropHandObject"

        if action == Action.OPEN_OBJECT.value:
            # The AI2-THOR Python library has buggy error checking
            # specifically for the OpenObject action,
            # so just use our own custom action here.
            return "MCSOpenObject"

        # if action == Action.ROTATE_OBJECT_IN_HAND.value:
        #     return "RotateHand"

        return action

    def retrieve_action_list_at_step(self, goal, step_number):
        return goal.retrieve_action_list_at_step(step_number)

    def retrieve_object_states(self, object_id):
        """Return the state list at the current step for the object with the
        given ID from the scene configuration data, if any."""
        self._scene_config.retrieve_object_states(
            object_id,
            self.__step_number)

    def stop_simulation(self):
        """Stop the 3D simulation environment. This controller won't work any
        more."""
        self._controller.stop()

    def _remove_none(self, d):
        '''Remove all none's from dictionaries'''
        for key, value in dict(d).items():
            if isinstance(value, dict):
                d[key] = self._remove_none(value)
            if isinstance(value, list):
                for index, val in enumerate(value):
                    if isinstance(val, dict):
                        value[index] = self._remove_none(val)
            if value is None:
                del d[key]
        return d

    def wrap_step(self, **kwargs):
        # whether or not to randomize segmentation mask colors
        consistentColors = False
        metadata_tier = self._config.get_metadata_tier()
        if(metadata_tier == ConfigManager.CONFIG_METADATA_TIER_ORACLE):
            consistentColors = True

        # Create the step data dict for the AI2-THOR step function.
        return dict(
            continuous=True,
            gridSize=self.GRID_SIZE,
            logs=True,
            renderDepthImage=self._config.is_depth_maps_enabled(),
            renderObjectImage=self._config.is_object_masks_enabled(),
            snapToGrid=False,
            consistentColors=consistentColors,
            **kwargs
        )

    def generate_noise(self):
        """
        Returns a random value between -0.05 and 0.05 used to add noise to all
        numerical action parameters noise_enabled is True.
        Returns
        -------
        float
            A value between -0.05 and 0.05 (using random.uniform).
        """

        return random.uniform(-0.5, 0.5)

    def get_metadata_level(self):
        """
        Returns the current metadata level set in the config. If none
        specified, returns 'default'.

        Returns
        -------
        string
            A string containing the current metadata level.
        """
        return self._config.get_metadata_tier()
